import torch
import torch.nn as nn
import copy
import torch.nn.functional as F
from typing import Dict
from torch_geometric.utils import to_dense_batch, to_dense_adj, dense_to_sparse
from torch_scatter import scatter
from esm import pretrained
from clipzyme.utils.classes import set_nox_type
from clipzyme.utils.registry import register_object, get_object
from clipzyme.utils.pyg import x_map
from clipzyme.models.abstract import AbstractModel
from clipzyme.models.chemprop import DMPNNEncoder


@register_object("enzyme_reaction_clip", "model")
class EnzymeReactionCLIP(AbstractModel):
    def __init__(self, args):
        super(EnzymeReactionCLIP, self).__init__()
        self.args = args
        self.reaction_clip_model_path = copy.copy(args.reaction_clip_model_path)
        self.use_as_protein_encoder = getattr(args, "use_as_protein_encoder", False)
        self.use_as_mol_encoder = getattr(args, "use_as_mol_encoder", False)
        args.train_esm_with_graph = getattr(args, "train_esm_with_graph", False)

        if args.reaction_clip_model_path is not None:
            state_dict = torch.load(args.reaction_clip_model_path)
            state_dict_copy = {
                k.replace("model.", "", 1): v
                for k, v in state_dict["state_dict"].items()
            }
            args = state_dict["hyper_parameters"]["args"]

        self.protein_encoder = get_object(args.protein_encoder, "model")(args)
        # option to train esm
        if args.train_esm_with_graph:
            self.esm_dir = args.train_esm_dir
            model, alphabet = pretrained.load_model_and_alphabet(args.train_esm_dir)
            self.esm_model = model
            self.alphabet = alphabet
            self.batch_converter = alphabet.get_batch_converter()

        self.ln_final = nn.LayerNorm(
            args.chemprop_hidden_dim
            if not args.use_protein_graphs
            else args.protein_dim
        )  # needs to be shape of protein_hidden, make it chemprop shape since we typically make these match

        self.logit_scale = nn.Parameter(
            torch.ones([]) * torch.log(torch.tensor(1 / 0.07))
        )

        wln_diff_args = copy.deepcopy(args)
        if args.model_name != "enzyme_reaction_clip_wldnv1":
            self.wln = DMPNNEncoder(args)  # WLN for mol representation
            wln_diff_args = copy.deepcopy(args)
            wln_diff_args.chemprop_edge_dim = args.chemprop_hidden_dim
            # wln_diff_args.chemprop_num_layers = 1
            self.wln_diff = DMPNNEncoder(wln_diff_args)

            # mol: attention pool
            self.final_linear = nn.Linear(
                args.chemprop_hidden_dim, args.chemprop_hidden_dim, bias=False
            )
            self.attention_fc = nn.Linear(args.chemprop_hidden_dim, 1, bias=False)

        # classifier
        if args.do_matching_task:
            self.mlp = get_object(args.mlp_name, "model")(args)

        if self.reaction_clip_model_path is not None:
            self.load_state_dict(state_dict_copy)

        if self.args.clip_freeze_esm:
            self.protein_encoder.requires_grad_(False)

    def encode_protein(self, batch):
        if self.args.use_protein_graphs:
            if self.args.train_esm_with_graph:
                sequences = [
                    (i, s) for i, s in enumerate(batch["graph"].structure_sequence)
                ]
                repr_layer = len(self.esm_model.layers)
                _, _, batch_tokens = self.batch_converter(sequences)
                batch_tokens = batch_tokens.to(self.logit_scale.device)
                mask = torch.logical_and(
                    torch.logical_and(
                        (batch_tokens != self.alphabet.cls_idx),
                        (batch_tokens != self.alphabet.eos_idx),
                    ),
                    (batch_tokens != self.alphabet.padding_idx),
                )
                out = self.esm_model(batch_tokens, repr_layers=[repr_layer])
                representations = out["representations"][repr_layer][mask]
                batch["graph"]["receptor"].x = representations
                if self.args.use_protein_msa:
                    msa = batch["graph"]["receptor"].msa
                    batch["graph"]["receptor"].x = torch.concat(
                        [representations, msa], dim=-1
                    )

            feats, coors = self.protein_encoder(batch)
            try:
                batch_idxs = batch["graph"]["receptor"].batch
            except:
                batch_idxs = batch["receptor"].batch
            protein_features = scatter(
                feats,
                batch_idxs,
                dim=0,
                reduce=self.args.pool_type,
            )
        else:
            if self.args.clip_freeze_esm:
                self.protein_encoder.requires_grad_(False)
                with torch.no_grad():
                    protein_features = self.protein_encoder(
                        {
                            "x": batch["sequence"],
                            "sequence": batch["sequence"],
                            "batch": batch,
                        }
                    )["hidden"]
            else:
                protein_features = self.protein_encoder(
                    {
                        "x": batch["sequence"],
                        "sequence": batch["sequence"],
                        "batch": batch,
                    }
                )["hidden"]

        # apply normalization
        protein_features = self.ln_final(protein_features)
        return protein_features

    def encode_reaction(self, batch):
        reactant_edge_feats = self.wln(batch["reactants"])[
            "edge_features"
        ]  # N x D, where N is all the nodes in the batch
        product_edge_feats = self.wln(batch["products"])[
            "edge_features"
        ]  # N x D, where N is all the nodes in the batch

        dense_reactant_edge_feats = to_dense_adj(
            edge_index=batch["reactants"].edge_index,
            edge_attr=reactant_edge_feats,
            batch=batch["reactants"].batch,
        )
        dense_product_edge_feats = to_dense_adj(
            edge_index=batch["products"].edge_index,
            edge_attr=product_edge_feats,
            batch=batch["products"].batch,
        )
        sum_vectors = dense_reactant_edge_feats + dense_product_edge_feats

        # undensify
        flat_sum_vectors = sum_vectors.sum(-1)
        new_edge_indices = [dense_to_sparse(E)[0] for E in flat_sum_vectors]
        new_edge_attr = torch.vstack(
            [sum_vectors[i, e[0], e[1]] for i, e in enumerate(new_edge_indices)]
        )
        cum_num_nodes = torch.cumsum(torch.bincount(batch["reactants"].batch), 0)
        new_edge_index = torch.hstack(
            [new_edge_indices[0]]
            + [ei + cum_num_nodes[i] for i, ei in enumerate(new_edge_indices[1:])]
        )
        reactants_and_products = batch["reactants"]
        reactants_and_products.edge_attr = new_edge_attr
        reactants_and_products.edge_index = new_edge_index

        # apply a separate WLN to the difference graph
        wln_diff_output = self.wln_diff(reactants_and_products)

        if self.args.aggregate_over_edges:
            edge_feats = wln_diff_output["edge_features"]
            edge_batch = batch["reactants"].batch[new_edge_index[0]]
            graph_feats = scatter(edge_feats, edge_batch, dim=0, reduce="sum")
        else:
            sum_node_feats = wln_diff_output["node_features"]
            sum_node_feats, _ = to_dense_batch(
                sum_node_feats, batch["products"].batch
            )  # num_candidates x max_num_nodes x D
            graph_feats = torch.sum(sum_node_feats, dim=-2)
        return graph_feats

    def forward(self, batch) -> Dict:
        output = {}
        if getattr(self.args, "use_as_protein_encoder", False):
            protein_features = self.encode_protein(batch)

            protein_features = protein_features / protein_features.norm(
                dim=1, keepdim=True
            )

            output.update(
                {
                    "hidden": protein_features,
                }
            )
            return output

        if getattr(self.args, "use_as_mol_encoder", False):
            substrate_features = self.encode_reaction(batch)
            substrate_features = substrate_features / substrate_features.norm(
                dim=1, keepdim=True
            )
            output.update(
                {
                    "hidden": substrate_features,
                }
            )
            return output

        substrate_features = self.encode_reaction(batch)
        protein_features = self.encode_protein(batch)

        # normalized features
        substrate_features = substrate_features / substrate_features.norm(
            dim=1, keepdim=True
        )
        protein_features = protein_features / protein_features.norm(dim=1, keepdim=True)

        output.update(
            {
                "substrate_hiddens": substrate_features,
                "protein_hiddens": protein_features,
            }
        )

        if self.args.do_matching_task:
            # take negatives based on EC
            ec = batch["ec2"] != batch["ec1"][:, None]
            ec = (ec / ec.sum(1))[:, None]
            neg_idx = []
            for e in ec:
                neg_idx.append(torch.multinomial(e, 1))
            neg_idx = torch.concat(neg_idx).squeeze()

            # take pairwise similarity of rxn embed and choose negatives
            # substrate_sim = substrate_features @ substrate_features.T
            # neg_idx = torch.argmin(substrate_sim, dim=-1)
            neg_samples = substrate_features[neg_idx]

            concat_hiddens_pos = torch.cat(
                [protein_features, substrate_features], dim=-1
            )
            concat_hiddens_neg = torch.cat([protein_features, neg_samples], dim=-1)
            concat_hiddens = torch.cat([concat_hiddens_pos, concat_hiddens_neg], dim=0)
            output["logit"] = self.mlp({"x": concat_hiddens})["logit"]
            bs = concat_hiddens_pos.shape[0]
            output["y"] = torch.cat(
                [concat_hiddens.new_ones(bs), concat_hiddens.new_zeros(bs)], dim=0
            )

        return output

    @staticmethod
    def add_args(parser) -> None:
        DMPNNEncoder.add_args(parser)
        parser.add_argument(
            "--mlp_name",
            type=str,
            action=set_nox_type("model"),
            default="mlp_classifier",
            help="Name of mlp to use",
        )
        parser.add_argument(
            "--do_matching_task",
            action="store_true",
            default=False,
            help="do molecule-protein matching",
        )
        parser.add_argument(
            "--protein_encoder",
            type=str,
            action=set_nox_type("model"),
            default="fair_esm2",
            help="Name of encoder to use",
        )
        parser.add_argument(
            "--reaction_clip_model_path",
            type=str,
            default=None,
            help="path to saved model if loading from pretrained",
        )
        parser.add_argument(
            "--aggregate_over_edges",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--clip_freeze_esm",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--use_as_protein_encoder",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--use_as_mol_encoder",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--train_esm_with_graph",
            action="store_true",
            default=False,
            help="train ESM model with graph NN.",
        )
        parser.add_argument(
            "--train_esm_dir",
            type=str,
            default="/home/snapshots/metabolomics/esm2/checkpoints/esm2_t33_650M_UR50D.pt",
            help="directory to load esm model from",
        )


@register_object("enzyme_reaction_clip_ec", "model")
class EnzymeReactionCLIPEC(EnzymeReactionCLIP):
    def __init__(self, args):
        super(EnzymeReactionCLIPEC, self).__init__(args)
        # the idea being that this protein_encoder produces both the protein and ec predictions
        # if pretrained ec model, load it
        if args.ec_model_model_path is not None:  # load pretrained model
            state_dict_ec = torch.load(args.ec_model_model_path)
            state_dict_ec_copy = {
                k.replace("model.", "", 1): v
                for k, v in state_dict_ec["state_dict"].items()
            }
            ec_args = state_dict_ec["hyper_parameters"]["args"]
            assert (
                args.protein_encoder == ec_args.protein_encoder
            ), "ec model name must match"
            assert (
                args.vocab_path == ec_args.vocab_path
            ), "pretrained model has different vocab"
            self.protein_encoder = get_object(args.protein_encoder, "model")(ec_args)
            self.protein_encoder.load_state_dict(state_dict_ec_copy)

    def encode_protein(self, batch):
        if self.args.use_protein_graphs and self.args.train_esm_with_graph:
            sequences = [
                (i, s) for i, s in enumerate(batch["graph"].structure_sequence)
            ]
            repr_layer = len(self.esm_model.layers)
            _, _, batch_tokens = self.batch_converter(sequences)
            batch_tokens = batch_tokens.to(self.logit_scale.device)
            mask = torch.logical_and(
                torch.logical_and(
                    (batch_tokens != self.alphabet.cls_idx),
                    (batch_tokens != self.alphabet.eos_idx),
                ),
                (batch_tokens != self.alphabet.padding_idx),
            )
            out = self.esm_model(batch_tokens, repr_layers=[repr_layer])
            representations = out["representations"][repr_layer][mask]
            batch["graph"]["receptor"].x = representations
            if self.args.use_protein_msa:
                msa = batch["graph"]["receptor"].msa
                batch["graph"]["receptor"].x = torch.concat(
                    [representations, msa], dim=-1
                )

        protein_output = self.protein_encoder(batch)
        # apply normalization
        protein_output["protein_features"] = self.ln_final(
            protein_output["protein_features"]
        )
        return protein_output

    def forward(self, batch) -> Dict:
        output = {}
        if getattr(self.args, "use_as_protein_encoder", False):
            encoded_protein_output = self.encode_protein(batch)
            protein_features = encoded_protein_output["protein_features"]

            protein_features = protein_features / protein_features.norm(
                dim=1, keepdim=True
            )

            output.update(
                {
                    "hidden": protein_features,
                }
            )
            return output

        if getattr(self.args, "use_as_mol_encoder", False):
            substrate_features = self.encode_reaction(batch)
            substrate_features = substrate_features / substrate_features.norm(
                dim=1, keepdim=True
            )
            output.update(
                {
                    "hidden": substrate_features,
                }
            )
            return output

        substrate_features = self.encode_reaction(batch)
        encoded_protein_output = self.encode_protein(batch)
        protein_features = encoded_protein_output["protein_features"]

        # normalized features
        substrate_features = substrate_features / substrate_features.norm(
            dim=1, keepdim=True
        )
        protein_features = protein_features / protein_features.norm(dim=1, keepdim=True)

        encoded_protein_output["protein_hiddens"] = protein_features
        encoded_protein_output["substrate_hiddens"] = substrate_features

        return encoded_protein_output

    @staticmethod
    def add_args(parser) -> None:
        DMPNNEncoder.add_args(parser)
        parser.add_argument(
            "--mlp_name",
            type=str,
            action=set_nox_type("model"),
            default="mlp_classifier",
            help="Name of mlp to use",
        )
        parser.add_argument(
            "--do_matching_task",
            action="store_true",
            default=False,
            help="do molecule-protein matching",
        )
        parser.add_argument(
            "--protein_encoder",
            type=str,
            action=set_nox_type("model"),
            default="fair_esm2",
            help="Name of encoder to use",
        )
        parser.add_argument(
            "--reaction_clip_model_path",
            type=str,
            default=None,
            help="path to saved model if loading from pretrained",
        )
        parser.add_argument(
            "--aggregate_over_edges",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--clip_freeze_esm",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--use_as_protein_encoder",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--use_as_mol_encoder",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--do_ec_task",
            action="store_true",
            default=False,
            help="do ec prediction",
        )
        parser.add_argument(
            "--ec_model_name",
            type=str,
            action=set_nox_type("model"),
            default="classifier",
            help="Name of model to use to predict ECs",
        )
        parser.add_argument(
            "--ec_model_model_path",
            type=str,
            default=None,
            help="path to saved model if loading from pretrained",
        )
        parser.add_argument(
            "--train_esm_with_graph",
            action="store_true",
            default=False,
            help="train ESM model with graph NN.",
        )


@register_object("enzyme_reaction_clip_cgr", "model")
class EnzymeReactionCLIPv2(EnzymeReactionCLIP):
    def encode_reaction(self, batch):
        dense_reactant_edge_feats = to_dense_adj(
            edge_index=batch["reactants"].edge_index,
            edge_attr=batch[
                "reactants"
            ].edge_attr,  # ! add 1 since encoding no edge here REMOVE with one-hot encoding
            batch=batch["reactants"].batch,
        )

        dense_product_edge_feats = to_dense_adj(
            edge_index=batch["products"].edge_index,
            edge_attr=batch[
                "products"
            ].edge_attr,  # ! add 1 since encoding no edge here
            batch=batch["products"].batch,
        )

        # node features
        # cgr_nodes = torch.cat([batch["reactants"].x, batch["products"].x], dim = -1)

        node_diff = batch["reactants"].x - batch["products"].x
        cgr_nodes = torch.cat(
            [batch["reactants"].x, node_diff[:, len(x_map["atomic_num"]) :]], dim=-1
        )

        # edge features
        # cgr_attr = torch.cat([dense_reactant_edge_feats, dense_product_edge_feats], dim = -1) # B, N, N, D
        cgr_attr = torch.cat(
            [
                dense_reactant_edge_feats,
                dense_reactant_edge_feats - dense_product_edge_feats,
            ],
            dim=-1,
        )

        # undensify
        flat_sum_vectors = cgr_attr.sum(-1)
        new_edge_indices = [dense_to_sparse(E)[0] for E in flat_sum_vectors]
        new_edge_attr = torch.vstack(
            [cgr_attr[i, e[0], e[1]] for i, e in enumerate(new_edge_indices)]
        )
        cum_num_nodes = torch.cumsum(torch.bincount(batch["reactants"].batch), 0)
        new_edge_index = torch.hstack(
            [new_edge_indices[0]]
            + [ei + cum_num_nodes[i] for i, ei in enumerate(new_edge_indices[1:])]
        )

        # make graph
        reactants_and_products = batch["reactants"]
        reactants_and_products.x = cgr_nodes
        reactants_and_products.edge_attr = new_edge_attr
        reactants_and_products.edge_index = new_edge_index

        # apply a separate WLN to the difference graph
        wln_diff_output = self.wln(reactants_and_products)

        if self.args.aggregate_over_edges:
            edge_feats = self.final_linear(wln_diff_output["edge_features"])
            edge_batch = batch["reactants"].batch[new_edge_index[0]]
            edge_feats, edge_mask = to_dense_batch(edge_feats, edge_batch)
            attn = self.attention_fc(edge_feats)
            attn[~edge_mask] = -torch.inf
            attn = torch.softmax(attn, -2)
            # graph_feats = scatter(edge_feats, edge_batch, dim = 0, reduce="sum")
            graph_feats = torch.sum(edge_feats * attn, dim=-2)
        else:
            node_feats = self.final_linear(wln_diff_output["node_features"])
            node_feats, node_mask = to_dense_batch(node_feats, batch["reactants"].batch)
            attn = self.attention_fc(node_feats)
            attn[~node_mask] = -torch.inf
            attn = torch.softmax(attn, -2)
            # node_feats, _ = to_dense_batch(node_feats, batch["reactants"].batch) # num_candidates x max_num_nodes x D
            graph_feats = torch.sum(node_feats * attn, dim=-2)
        return graph_feats


@register_object("enzyme_reaction_clip_string", "model")
class EnzymeReactionCLIPString(EnzymeReactionCLIP):
    def __init__(self, args):
        super(EnzymeReactionCLIPString, self).__init__(args)
        self.substrate_encoder = get_object(args.substrate_encoder, "model")(args)

    def encode_reaction(self, batch):
        feats = self.substrate_encoder(batch)["encoder_output"]
        return feats

    @staticmethod
    def add_args(parser) -> None:
        super(EnzymeReactionCLIPString, EnzymeReactionCLIPString).add_args(parser)
        parser.add_argument(
            "--substrate_encoder",
            type=str,
            action=set_nox_type("model"),
            default="full_reaction_encoder",
            help="Rxn String Encoder",
        )


@register_object("enzyme_reaction_clip_pretrained", "model")
class EnzymeReactionCLIPPretrained(EnzymeReactionCLIP):
    def __init__(self, args):
        super(EnzymeReactionCLIPPretrained, self).__init__(args)
        self.substrate_encoder = get_object(args.substrate_encoder, "model")(args)
        if args.substrate_model_path is not None:
            state_dict = torch.load(args.substrate_model_path)
            state_dict_copy = {
                k.replace("model.", "", 1): v
                for k, v in state_dict["state_dict"].items()
            }
            # Remove keys from state_dict that are not in the model
            model_state_dict_keys = set(self.substrate_encoder.state_dict().keys())
            state_dict_keys = list(
                state_dict_copy.keys()
            )  # We use list to avoid RuntimeError for changing dict size during iteration
            for key in state_dict_keys:
                if key not in model_state_dict_keys:
                    del state_dict_copy[key]
            self.substrate_encoder.load_state_dict(state_dict_copy)

        if args.chemprop_hidden_dim != args.protein_dim:
            self.substrate_projection = nn.Linear(
                args.chemprop_hidden_dim, args.protein_dim
            )  # needs to be shape of protein_hidden, make it chemprop shape since we typically make these match

    def encode_reaction(self, batch):
        feats = self.substrate_encoder(batch)
        # unbatch the graph
        feats, mask = to_dense_batch(feats["c_final"], batch=batch["mol"].batch)
        feats = feats.sum(1)  # sum over all nodes
        feats = self.substrate_projection(feats)
        return feats

    @staticmethod
    def add_args(parser) -> None:
        super(EnzymeReactionCLIPPretrained, EnzymeReactionCLIPPretrained).add_args(
            parser
        )
        parser.add_argument(
            "--substrate_encoder",
            type=str,
            action=set_nox_type("model"),
            default="full_reaction_encoder",
            help="Rxn String Encoder",
        )
        parser.add_argument(
            "--substrate_model_path",
            type=str,
            default=None,
            help="path to saved model if loading from pretrained",
        )


@register_object("enzyme_reaction_clip_wldn", "model")
class EnzymeReactionCLIPWLDN(EnzymeReactionCLIPPretrained):
    def __init__(self, args):
        super(EnzymeReactionCLIPWLDN, self).__init__(args)
        del self.substrate_encoder.reactivity_net

    def encode_reaction(self, batch):
        gat_output = self.substrate_encoder.wln(batch["reactants"])
        reactant_node_feats = gat_output["node_features"]

        gat_output = self.substrate_encoder.wln(batch["products"])
        product_node_feats = gat_output["node_features"]

        difference_vectors = product_node_feats - reactant_node_feats

        product_graph = batch["products"].clone()
        product_graph.x = difference_vectors

        # apply a separate WLN to the difference graph
        wln_diff_output = self.substrate_encoder.wln_diff(product_graph)
        diff_node_feats = wln_diff_output["node_features"]
        diff_node_feats, mask = to_dense_batch(
            diff_node_feats, batch=product_graph.batch
        )
        feats = diff_node_feats.sum(1)  # sum over all nodes
        if feats.shape[-1] != self.args.protein_dim:
            feats = self.substrate_projection(feats)
        return feats

    def add_args(parser) -> None:
        parser.add_argument(
            "--substrate_model_path",
            type=str,
            default=None,
            help="path to saved model if loading from pretrained",
        )
        parser.add_argument(
            "--substrate_encoder",
            type=str,
            action=set_nox_type("model"),
            default="full_reaction_encoder",
            help="Rxn String Encoder",
        )
        parser.add_argument(
            "--mlp_name",
            type=str,
            action=set_nox_type("model"),
            default="mlp_classifier",
            help="Name of mlp to use",
        )
        parser.add_argument(
            "--do_matching_task",
            action="store_true",
            default=False,
            help="do molecule-protein matching",
        )
        parser.add_argument(
            "--protein_encoder",
            type=str,
            action=set_nox_type("model"),
            default="fair_esm2",
            help="Name of encoder to use",
        )
        parser.add_argument(
            "--reaction_clip_model_path",
            type=str,
            default=None,
            help="path to saved model if loading from pretrained",
        )
        parser.add_argument(
            "--aggregate_over_edges",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--clip_freeze_esm",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--use_as_protein_encoder",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--use_as_mol_encoder",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--train_esm_with_graph",
            action="store_true",
            default=False,
            help="train ESM model with graph NN.",
        )
        parser.add_argument(
            "--train_esm_dir",
            type=str,
            default="/home/snapshots/metabolomics/esm2/checkpoints/esm2_t33_650M_UR50D.pt",
            help="directory to load esm model from",
        )



"""
Adapted from: https://github.com/itakigawa/pyg_chemprop/blob/main/pyg_chemprop.py
"""

import torch
import torch.nn as nn
from torch_geometric.nn import global_mean_pool
from torch_scatter import scatter_sum
from tqdm import tqdm
from clipzyme.utils.registry import register_object
from clipzyme.models.abstract import AbstractModel


def get_reverse_edge_indices(edge_index):
    revedge_index = torch.zeros(edge_index.shape[1]).long()
    for k, (i, j) in enumerate(zip(*edge_index)):
        edge_to_i = edge_index[1] == i
        edge_from_j = edge_index[0] == j
        revedge_index[k] = torch.where(edge_to_i & edge_from_j)[0]
    return revedge_index


def directed_mp(message, edge_index, revedge_index):
    m = scatter_sum(message, edge_index[1], dim=0)  # all messages going to edge i
    m_all = m[edge_index[0]]
    m_rev = message[revedge_index]  # reverse message for i -> j at index i
    return m_all - m_rev


def aggregate_at_nodes(num_nodes, message, edge_index):
    m = scatter_sum(message, edge_index[1], dim=0, dim_size=num_nodes)
    return m[torch.arange(num_nodes)]


@register_object("chemprop", "model")
class DMPNNEncoder(AbstractModel):
    def __init__(self, args):
        super(DMPNNEncoder, self).__init__()
        hidden_size = args.chemprop_hidden_dim
        node_fdim = args.chemprop_node_dim
        edge_fdim = args.chemprop_edge_dim

        self.act_func = nn.ReLU()
        self.W1 = nn.Linear(node_fdim + edge_fdim, hidden_size, bias=False)
        self.W2 = nn.Linear(hidden_size, hidden_size, bias=False)
        self.W3 = nn.Linear(node_fdim + hidden_size, hidden_size, bias=True)
        self.depth = args.chemprop_num_layers

        self.pool_type = args.chemprop_pool

    def forward(self, data):
        x, edge_index, edge_attr, num_nodes, batch = (
            data.x,
            data.edge_index,
            data.edge_attr,
            data.num_nodes,
            data.batch,
        )

        # get indices of reverse direction
        revedge_index = get_reverse_edge_indices(edge_index)

        # initialize messages on edges
        init_msg = torch.cat([x[edge_index[0]], edge_attr], dim=1).float()
        h0 = self.act_func(self.W1(init_msg))

        # directed message passing over edges
        h = h0
        for _ in range(self.depth - 1):
            m = directed_mp(h, edge_index, revedge_index)
            h = self.act_func(h0 + self.W2(m))

        # aggregate in-edge messages at nodes
        v_msg = aggregate_at_nodes(num_nodes, h, edge_index)

        z = torch.cat([x, v_msg], dim=1)
        node_attr = self.act_func(self.W3(z))

        # readout: pyg global pooling
        output = {}
        output["node_features"] = node_attr
        output["edge_features"] = h  # for every pair (i,j) as in edge_index
        if self.pool_type != "none":
            output["graph_features"] = global_mean_pool(node_attr, batch)
            output["hidden"] = output["graph_features"]

        return output

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--chemprop_num_layers",
            type=int,
            default=1,
            help="Number of layers in GNN, equivalently number of convolution iterations.",
        )
        parser.add_argument(
            "--chemprop_hidden_dim",
            type=int,
            default=None,
            help="Dimension of hidden layers (node features)",
        )
        parser.add_argument(
            "--chemprop_node_dim",
            type=int,
            default=None,
            help="Node feature dimensionality.",
        )
        parser.add_argument(
            "--chemprop_edge_dim",
            type=int,
            default=None,
            help="Edge feature dimensionality (in case there are any).",
        )
        parser.add_argument(
            "--chemprop_pool",
            type=str,
            choices=["none", "sum", "mul", "mean", "min", "max"],
            default="sum",
            help="Type of pooling to do to obtain graph features",
        )


@register_object("wln_encoder", "model")
class WLNEncoder(AbstractModel):
    def __init__(self, args):
        super(WLNEncoder, self).__init__()
        hidden_size = args.wln_enc_hidden_dim
        node_fdim = args.wln_enc_node_dim
        edge_fdim = args.wln_enc_edge_dim

        self.act_func = nn.ReLU()
        self.W0 = nn.Linear(node_fdim, hidden_size, bias=False)
        self.W1 = nn.Linear(hidden_size + edge_fdim, hidden_size, bias=False)
        self.W2 = nn.Linear(2 * hidden_size, hidden_size, bias=True)

        self.depth = args.wln_enc_num_layers
        self.pool_type = args.wln_enc_pool

    def forward(self, data):
        output = {}
        x, edge_index, edge_attr, num_nodes, batch = (
            data.x,
            data.edge_index,
            data.edge_attr,
            data.num_nodes,
            data.batch,
        )

        # project node features
        node_feats = self.act_func(self.W0(x.float()))

        for _ in range(self.depth):
            # concatenate source node features with edge features
            concat_edge_feats = torch.cat(
                [node_feats[edge_index[0]], edge_attr], dim=1
            ).float()
            # project
            he = self.act_func(self.W1(concat_edge_feats))
            # message passing of edge features into nodes
            hv_new = aggregate_at_nodes(num_nodes, he, edge_index)
            # update nodes
            node_feats = self.act_func(self.W2(torch.cat([node_feats, hv_new], dim=1)))

        output["node_features"] = node_feats
        output["edge_features"] = he
        if self.pool_type != "none":
            output["graph_features"] = global_mean_pool(node_feats, batch)
            output["hidden"] = output["graph_features"]
        return output

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--wln_enc_num_layers",
            type=int,
            default=1,
            help="Number of layers in GNN, equivalently number of convolution iterations.",
        )
        parser.add_argument(
            "--wln_enc_hidden_dim",
            type=int,
            default=None,
            help="Dimension of hidden layers (node features)",
        )
        parser.add_argument(
            "--wln_enc_node_dim",
            type=int,
            default=None,
            help="Node feature dimensionality.",
        )
        parser.add_argument(
            "--wln_enc_edge_dim",
            type=int,
            default=None,
            help="Edge feature dimensionality (in case there are any).",
        )
        parser.add_argument(
            "--wln_enc_pool",
            type=str,
            choices=["none", "sum", "mul", "mean", "min", "max"],
            default="sum",
            help="Type of pooling to do to obtain graph features",
        )



"""
Copyright (c) 2022, salesforce.com, inc.
All rights reserved.
SPDX-License-Identifier: BSD-3-Clause
For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

import math
from clipzyme.utils.registry import register_object
from clipzyme.utils.classes import Nox
from torch.optim.lr_scheduler import _LRScheduler


@register_object("linear_warmup_step_lr", "scheduler")
class LinearWarmupStepLRScheduler(_LRScheduler, Nox):
    def __init__(self, optimizer, args):
        max_epoch = args.max_epochs
        min_lr = args.warmup_min_lr
        init_lr = args.warmup_init_lr
        decay_rate = args.warmup_decay_rate
        warmup_start_lr = args.warmup_start_lr
        warmup_steps = args.warmup_steps

        self.optimizer = optimizer

        self.max_epoch = max_epoch
        self.min_lr = min_lr

        self.decay_rate = decay_rate

        self.init_lr = init_lr
        self.warmup_steps = warmup_steps
        self.warmup_start_lr = warmup_start_lr if warmup_start_lr >= 0 else init_lr
        super().__init__(optimizer)

    def step(self, cur_epoch=0):
        self._step_count += 1
        cur_step = self._step_count
        if cur_step < self.warmup_steps:
            warmup_lr_schedule(
                step=cur_step,
                optimizer=self.optimizer,
                max_step=self.warmup_steps,
                init_lr=self.warmup_start_lr,
                max_lr=self.init_lr,
            )
        else:
            step_lr_schedule(
                epoch=cur_epoch,
                optimizer=self.optimizer,
                init_lr=self.init_lr,
                min_lr=self.min_lr,
                decay_rate=self.decay_rate,
            )

    @staticmethod
    def add_args(parser) -> None:
        parser.add_argument(
            "--warmup_min_lr",
            type=float,
            default=None,
            help="minimum learning rate",
        )
        parser.add_argument(
            "--warmup_init_lr",
            type=float,
            default=None,
            help="base learning rate",
        )
        parser.add_argument(
            "--warmup_decay_rate",
            type=float,
            default=1,
            help="decay rate",
        )
        parser.add_argument(
            "--warmup_start_lr",
            type=float,
            default=-1,
            help="lr of warmup",
        )
        parser.add_argument(
            "--warmup_steps",
            type=float,
            default=0,
            help="step of warmpu",
        )


@register_object("linear_warmup_cosine_lr", "scheduler")
class LinearWarmupCosineLRScheduler(_LRScheduler, Nox):
    def __init__(self, optimizer, args):
        max_epoch = args.max_epochs
        min_lr = args.warmup_min_lr
        init_lr = args.warmup_init_lr
        warmup_start_lr = args.warmup_start_lr
        warmup_steps = args.warmup_steps

        self.optimizer = optimizer

        self.max_epoch = max_epoch
        self.min_lr = min_lr

        self.init_lr = init_lr
        self.warmup_steps = warmup_steps
        self.warmup_start_lr = warmup_start_lr if warmup_start_lr >= 0 else init_lr
        super().__init__(optimizer)

    def step(self, cur_epoch=0):
        # assuming the warmup iters less than one epoch
        self._step_count += 1
        cur_step = self._step_count
        if cur_step < self.warmup_steps:
            warmup_lr_schedule(
                step=cur_step,
                optimizer=self.optimizer,
                max_step=self.warmup_steps,
                init_lr=self.warmup_start_lr,
                max_lr=self.init_lr,
            )
        else:
            cosine_lr_schedule(
                epoch=cur_epoch,
                optimizer=self.optimizer,
                max_epoch=self.max_epoch,
                init_lr=self.init_lr,
                min_lr=self.min_lr,
            )

    @staticmethod
    def add_args(parser) -> None:
        parser.add_argument(
            "--warmup_min_lr",
            type=float,
            default=None,
            help="minimum learning rate",
        )
        parser.add_argument(
            "--warmup_init_lr",
            type=float,
            default=None,
            help="base learning rate",
        )
        parser.add_argument(
            "--warmup_start_lr",
            type=float,
            default=-1,
            help="lr of warmup",
        )
        parser.add_argument(
            "--warmup_steps",
            type=float,
            default=0,
            help="step of warmpu",
        )


@register_object("constant_lr", "scheduler")
class ConstantLRScheduler(_LRScheduler, Nox):
    def __init__(self, optimizer, args):
        init_lr = args.init_lr
        warmup_start_lr = args.warmup_start_lr
        warmup_steps = args.warmup_steps

        self.optimizer = optimizer
        self.lr = init_lr
        self.warmup_start_lr = warmup_start_lr if warmup_start_lr >= 0 else init_lr
        self.warmup_steps = warmup_steps
        super().__init__(optimizer)

    def step(self, cur_epoch=0):
        cur_step = self._step_count

        if self.warmup_steps < cur_step:
            warmup_lr_schedule(
                step=cur_step,
                optimizer=self.optimizer,
                max_step=self.warmup_steps,
                init_lr=self.warmup_start_lr,
                max_lr=self.lr,
            )
        else:
            for param_group in self.optimizer.param_groups:
                param_group["lr"] = self.lr

    @staticmethod
    def add_args(parser) -> None:
        parser.add_argument(
            "--warmup_init_lr",
            type=float,
            default=None,
            help="base learning rate",
        )
        parser.add_argument(
            "--warmup_start_lr",
            type=float,
            default=-1,
            help="lr of warmup",
        )
        parser.add_argument(
            "--warmup_steps",
            type=float,
            default=0,
            help="step of warmpu",
        )


def cosine_lr_schedule(optimizer, epoch, max_epoch, init_lr, min_lr):
    """Decay the learning rate"""
    lr = (init_lr - min_lr) * 0.5 * (
        1.0 + math.cos(math.pi * epoch / max_epoch)
    ) + min_lr
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def warmup_lr_schedule(optimizer, step, max_step, init_lr, max_lr):
    """Warmup the learning rate"""
    lr = min(max_lr, init_lr + (max_lr - init_lr) * step / max(max_step, 1))
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def step_lr_schedule(optimizer, epoch, init_lr, min_lr, decay_rate):
    """Decay the learning rate"""
    lr = max(min_lr, init_lr * (decay_rate**epoch))
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr



from pytorch_lightning.callbacks import Callback
from clipzyme.utils.registry import get_object


def set_callbacks(trainer, args):
    """
    Set callbacks for trainer, taking into consideration callbacks already set by trainer args.
    Callbacks that are preset by args and perform the same function as those manually selected
    are removed by comparing parent classes between callbacks.

    Parameters
    ----------
    trainer : pl.Trainer
        lightning trainer
    args : Namespace
        global args

    Returns
    -------
    callbacks: list
        complete list of callbacks to be used by trainer
    """
    callbacks = []
    for cback in args.callback_names:
        callbacks.append(get_object(cback, "callback")(args))

    # remove callbacks that are set manually
    redundant_callbacks = []
    for cback in trainer.callbacks:
        parent_cls_preset = get_callback_parent_class(cback)
        for new_cback in callbacks:
            parent_cls_postset = get_callback_parent_class(new_cback)
            if parent_cls_preset == parent_cls_postset:
                redundant_callbacks.append(cback)

    for cback in trainer.callbacks:
        if cback not in redundant_callbacks:
            callbacks.append(cback)

    return callbacks


def get_callback_parent_class(obj):
    """
    Parameters
    ----------
    obj : Callback
        instance of a callback class

    Returns
    -------
    class
        parent class of callback that is the first child of the Callback class
    """
    parent_id = [cls == Callback for cls in obj.__class__.__mro__].index(True)
    parent_cls = obj.__class__.__mro__[parent_id - 1]
    return parent_cls



import traceback, warnings
import argparse
from typing import List, Literal
from abc import ABCMeta, abstractmethod
import json
import pickle
import os
from collections import Counter
import numpy as np
from torch.utils import data
from clipzyme.utils.classes import Nox, set_nox_type, classproperty
from clipzyme.utils.messages import METAFILE_NOTFOUND_ERR, LOAD_FAIL_MSG
from rich import print as rprint


class AbstractDataset(data.Dataset, Nox):
    def __init__(self, args: argparse.ArgumentParser, split_group: str) -> None:
        """
        Abstract Dataset
        params: args - config.
        params: split_group - ['train'|'dev'|'test'].

        constructs: standard pytorch Dataset obj, which can be fed in a DataLoader for batching

        # 1. Load dataset and optionally item loaders
        # 2. Create dataset
        # 3. Assign splits based on processed data
        # 4. Post-process dataset
        # 5. Create dataset for specific split group
        # 6. Optionally set sample weights
        # 7. Print summary statement
        """
        __metaclass__ = ABCMeta

        super(AbstractDataset, self).__init__()

        self.split_group = split_group
        self.args = args

        self.init_class(args, split_group)

        if args.dataset_cache_path is not None:
            try:
                self.dataset = pickle.load(open(args.dataset_cache_path, "rb"))
            except:
                self.dataset = self.create_dataset(split_group)
                pickle.dump(self.dataset, open(args.dataset_cache_path, "wb"))
        else:
            self.dataset = self.create_dataset(split_group)

        if len(self.dataset) == 0:
            return

        if args.assign_splits:
            self.assign_splits(self.dataset, args.split_probs, args.split_seed)
        else:
            rprint("[magenta]WARNING: `assign_splits` = False[/magenta]")

        self.post_process(args)

        self.dataset = self.get_split_group_dataset(self.dataset, split_group)

        self.set_sample_weights(args)

        self.print_summary_statement(self.dataset, split_group)

    def init_class(self, args: argparse.ArgumentParser, split_group: str) -> None:
        """Perform Class-Specific init methods
           Default is to load JSON dataset

        Args:
            args (argparse.ArgumentParser)
            split_group (str)
        """
        self.load_dataset(args)

    def post_process(self, args):
        """Perform post-processing on dataset
        This may include computing statistics based on training set

        Args:
            args (argparse.ArgumentParser)
        """
        pass

    def load_dataset(self, args: argparse.ArgumentParser) -> None:
        """Loads dataset file

        Args:
            args (argparse.ArgumentParser)

        Raises:
            Exception: Unable to load
        """
        try:
            self.metadata_json = json.load(open(args.dataset_file_path, "r"))
        except Exception as e:
            raise Exception(METAFILE_NOTFOUND_ERR.format(args.dataset_file_path, e))

    @abstractmethod
    def create_dataset(
        self, split_group: Literal["train", "dev", "test"]
    ) -> List[dict]:
        """
        Creates the dataset of samples from json metadata file.
        """
        pass

    def get_split_group_dataset(
        self, processed_dataset, split_group: Literal["train", "dev", "test"]
    ) -> List[dict]:
        """Get split group dataset"""
        return processed_dataset

    @abstractmethod
    def skip_sample(self, sample) -> bool:
        """
        Return True if sample should be skipped and not included in data
        """
        return False

    @abstractmethod
    def check_label(self, sample) -> bool:
        """
        Return True if the row contains a valid label for the task
        """
        pass

    @abstractmethod
    def get_label(self, sample):
        """
        Get task specific label for a given sample
        """
        pass

    @property
    @abstractmethod
    def SUMMARY_STATEMENT(self) -> None:
        """
        Prints summary statement with dataset stats
        """
        pass

    def print_summary_statement(self, dataset, split_group):
        statement = "{} DATASET CREATED FOR {}.{}".format(
            split_group.upper(), self.args.dataset_name.upper(), self.SUMMARY_STATEMENT
        )
        print(statement)

    def __len__(self) -> int:
        return len(self.dataset)

    @abstractmethod
    def __getitem__(self, index):
        """
        Fetch single sample from dataset

        Args:
            index (int): random index of sample from dataset

        Returns:
            sample (dict): a sample
        """
        sample = self.dataset[index]
        try:
            return sample
        except Exception:
            warnings.warn(
                LOAD_FAIL_MSG.format(sample["sample_id"], traceback.print_exc())
            )

    def assign_splits(self, metadata_json, split_probs, seed=0) -> None:
        """
        Assign samples to data splits

        Args:
            metadata_json (dict): raw json dataset loaded
        """
        np.random.seed(seed)
        for idx in range(len(metadata_json)):
            metadata_json[idx]["split"] = np.random.choice(
                ["train", "dev", "test"], p=split_probs
            )

    def set_sample_weights(self, args: argparse.ArgumentParser) -> None:
        """
        Set weights for each sample

        Args:
            args (argparse.ArgumentParser)
        """
        if args.class_bal:
            label_dist = [d[args.class_bal_key] for d in self.dataset]
            label_counts = Counter(label_dist)
            weight_per_label = 1.0 / len(label_counts)
            label_weights = {
                label: weight_per_label / count for label, count in label_counts.items()
            }

            print("Class counts are: {}".format(label_counts))
            print("Label weights are {}".format(label_weights))
            self.weights = [label_weights[d[args.class_bal_key]] for d in self.dataset]
        else:
            pass

    @classproperty
    def DATASET_ITEM_KEYS(cls) -> list:
        """
        List of keys to be included in sample when being batched

        Returns:
            list
        """
        standard = ["sample_id"]
        return standard

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--class_bal", action="store_true", default=False, help="class balance"
        )
        parser.add_argument(
            "--class_bal_key",
            type=str,
            default="y",
            help="dataset key to use for class balancing",
        )
        parser.add_argument(
            "--dataset_file_path",
            type=str,
            default=None,
            help="Path to dataset file",
        )
        parser.add_argument(
            "--data_dir",
            type=str,
            default=None,
            help="Path to dataset directory",
        )
        parser.add_argument(
            "--num_classes", type=int, default=6, help="Number of classes to predict"
        )
        # Alternative training/testing schemes
        parser.add_argument(
            "--assign_splits",
            action="store_true",
            default=False,
            help="Whether to assign different splits than those predetermined in dataset",
        )
        parser.add_argument(
            "--split_type",
            type=str,
            default="random",
            help="How to split dataset if assign_split = True..",
        )
        parser.add_argument(
            "--split_probs",
            type=float,
            nargs="+",
            default=[0.6, 0.2, 0.2],
            help="Split probs for datasets without fixed train dev test. ",
        )
        parser.add_argument(
            "--split_seed",
            type=int,
            default=0,
            help="seed for consistent randomization",
        )
        # loader
        parser.add_argument(
            "--input_loader_name",
            type=str,
            action=set_nox_type("input_loader"),
            default=None,
            help="input loader",
        )
        # cache dataset
        parser.add_argument(
            "--dataset_cache_path",
            type=str,
            default=None,
            help="path to dataset cache",
        )



import torch
from typing import List, Dict
from clipzyme.models.abstract import AbstractModel
from clipzyme.utils.registry import register_object
from clipzyme.utils.smiles import standardize_reaction, tokenize_smiles
from transformers import (
    BertConfig,
    BertModel,
    BertTokenizer,
)


@register_object("full_reaction_encoder", "model")
class FullReactionEncoder(AbstractModel):
    def __init__(self, args):
        super().__init__()

        self.tokenizer = self.init_tokenizer(args)

        enc_config = BertConfig(
            max_position_embeddings=args.max_seq_len,
            vocab_size=self.tokenizer.vocab_size,
            output_hidden_states=True,
            output_attentions=True,
            hidden_size=args.hidden_size,
            intermediate_size=args.intermediate_size,
            embedding_size=args.embedding_size,
            num_attention_heads=args.num_heads,
            num_hidden_layers=args.num_hidden_layers,
        )

        self.model = BertModel(enc_config)
        self.config = self.model.config
        self.args = args
        self.register_buffer("devicevar", torch.zeros(1, dtype=torch.int8))

    @staticmethod
    def init_tokenizer(args):
        return BertTokenizer(
            vocab_file=args.vocab_path,
            do_lower_case=False,
            do_basic_tokenize=False,
            # sep_token=".",
            padding=True,
            truncation=True,
            model_max_length=args.max_seq_len,
            cls_token="[CLS]",
            eos_token="[EOS]",
        )

    @staticmethod
    def tokenize(list_of_text: List[str], tokenizer, args):
        # standardize reaction
        x = [standardize_reaction(r) for r in list_of_text]
        x = [tokenize_smiles(r, return_as_str=True) for r in x]

        if args.use_cls_token:
            # add [CLS] and [EOS] tokens
            x = [f"{tokenizer.cls_token} {r} {tokenizer.eos_token}" for r in x]

        # tokenize str characters into tensor of indices with corresponding masks
        tokenized_inputs = tokenizer(
            x,
            add_special_tokens=False,
            padding=True,
            truncation=True,
            max_length=tokenizer.model_max_length,
            return_tensors="pt",
        )

        # get mask for special tokens that are not masked in MLM (return_special_tokens_mask=True doesn't work for additional special tokens)
        special_tokens_mask = [
            tokenizer.get_special_tokens_mask(
                toks, None, already_has_special_tokens=True
            )
            for toks in tokenized_inputs["input_ids"]
        ]

        tokenized_inputs["special_tokens_mask"] = torch.tensor(
            special_tokens_mask, dtype=torch.int64
        )

        return tokenized_inputs

    def forward(self, batch) -> Dict:
        # tokenize full reaction
        encoder_input_ids = self.tokenize(batch["reaction"], self.tokenizer, self.args)

        # move to device
        for k, v in encoder_input_ids.items():
            encoder_input_ids[k] = v.to(self.devicevar.device)

        return_dict = self.config.use_return_dict

        encoder_outputs = self.model(
            input_ids=encoder_input_ids["input_ids"],
            attention_mask=encoder_input_ids["attention_mask"],
        )

        encoder_hidden_states = encoder_outputs[0]

        output = {
            "encoder_output": encoder_outputs["pooler_output"]
            # "encoder_hidden_states": encoder_outputs.hidden_states,
            # "encoder_attentions": encoder_outputs.attentions,
        }
        return output

    @staticmethod
    def add_args(parser) -> None:
        parser.add_argument(
            "--vocab_path",
            type=str,
            default=None,
            required=True,
            help="path to vocab text file required for tokenizer",
        )
        parser.add_argument(
            "--merges_file_path",
            type=str,
            default=None,
            help="path to merges file required for RoBerta tokenizer",
        )
        parser.add_argument(
            "--num_hidden_layers",
            type=int,
            default=6,
            help="number of layers in the transformer",
        )
        parser.add_argument(
            "--max_seq_len",
            type=int,
            default=512,
            help="maximum length allowed for the input sequence",
        )
        parser.add_argument(
            "--hidden_size",
            type=int,
            default=256,
            help="maximum length allowed for the input sequence",
        )
        parser.add_argument(
            "--intermediate_size",
            type=int,
            default=512,
            help="maximum length allowed for the input sequence",
        )
        parser.add_argument(
            "--embedding_size",
            type=int,
            default=128,
            help="maximum length allowed for the input sequence",
        )
        parser.add_argument(
            "--num_heads",
            type=int,
            default=8,
            help="maximum length allowed for the input sequence",
        )
        parser.add_argument(
            "--use_cls_token",
            action="store_true",
            default=False,
            help="use cls token as hidden representation of sequence",
        )



import torch
import torch.nn as nn
import copy
from clipzyme.utils.registry import register_object, get_object
from clipzyme.utils.classes import set_nox_type
from clipzyme.models.abstract import AbstractModel


@register_object("classifier", "model")
class Classifier(AbstractModel):
    def __init__(self, args):
        super(Classifier, self).__init__()

        self.args = args
        self.logit_scale = nn.Parameter(torch.ones([]) * torch.log(torch.tensor(1 / 2)))
        if args.pretrained_encoder_path:
            state_dict = torch.load(args.pretrained_encoder_path)
            self.encoder = get_object(args.model_name_for_encoder, "model")(
                state_dict["hyper_parameters"]["args"]
            )
            self.encoder.load_state_dict(
                {k[len("model.") :]: v for k, v in state_dict["state_dict"].items()}
            )
            self.encoder.args = args
        else:
            self.encoder = get_object(args.model_name_for_encoder, "model")(args)

        cargs = copy.deepcopy(args)
        cargs.mlp_input_dim = args.classifier_mlp_input_dim
        cargs.mlp_layer_configuration = args.classifier_mlp_layer_configuration
        cargs.mlp_use_batch_norm = args.classifier_mlp_use_batch_norm
        cargs.mlp_use_layer_norm = args.classifier_mlp_use_layer_norm
        self.mlp = get_object("mlp_classifier", "model")(cargs)
        self.encoder_hidden_key = args.encoder_hidden_key

    def forward(self, batch=None):
        output = {}
        output["encoder_hidden"] = self.encoder(batch)[self.encoder_hidden_key]
        output.update(self.mlp({"x": output["encoder_hidden"]}))
        return output

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--model_name_for_encoder",
            type=str,
            action=set_nox_type("model"),
            default="resnet18",
            help="Name of encoder to use",
        )
        parser.add_argument(
            "--pretrained_encoder_path",
            type=str,
            default=None,
            help="path to pretrained if applicable",
        )
        parser.add_argument(
            "--encoder_hidden_key",
            type=str,
            default="hidden",
            help="name of hidden features from encoder output",
        )
        parser.add_argument(
            "--classifier_mlp_input_dim",
            type=int,
            default=512,
            help="Dim of input to mlp",
        )
        parser.add_argument(
            "--classifier_mlp_layer_configuration",
            type=int,
            nargs="*",
            default=[],
            help="MLP layer dimensions",
        )
        parser.add_argument(
            "--classifier_mlp_use_batch_norm",
            action="store_true",
            default=False,
            help="Use batchnorm in mlp",
        )
        parser.add_argument(
            "--classifier_mlp_use_layer_norm",
            action="store_true",
            default=False,
            help="Use LayerNorm in mlp",
        )


@register_object("mlp_classifier", "model")
class MLPClassifier(AbstractModel):
    def __init__(self, args):
        super(MLPClassifier, self).__init__()

        self.args = args

        model_layers = []
        cur_dim = args.mlp_input_dim
        for layer_size in args.mlp_layer_configuration:
            model_layers.extend(self.append_layer(cur_dim, layer_size, args))
            cur_dim = layer_size

        self.mlp = nn.Sequential(*model_layers)
        self.predictor = nn.Linear(cur_dim, args.num_classes)

    def append_layer(self, cur_dim, layer_size, args, with_dropout=True):
        linear_layer = nn.Linear(cur_dim, layer_size)
        bn = nn.BatchNorm1d(layer_size)
        ln = nn.LayerNorm(layer_size)
        if args.mlp_use_batch_norm:
            seq = [linear_layer, bn, nn.ReLU()]
        elif args.mlp_use_layer_norm:
            seq = [linear_layer, ln, nn.ReLU()]
        else:
            seq = [linear_layer, nn.ReLU()]
        if with_dropout:
            seq.append(nn.Dropout(p=args.dropout))
        return seq

    def forward(self, batch=None):
        output = {}
        z = self.mlp(batch["x"])
        output["hidden"] = z
        output["logit"] = self.predictor(z)
        return output

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            "--mlp_input_dim", type=int, default=512, help="Dim of input to mlp"
        )
        parser.add_argument(
            "--mlp_layer_configuration",
            type=int,
            nargs="*",
            default=[],
            help="MLP layer dimensions",
        )
        parser.add_argument(
            "--mlp_use_batch_norm",
            action="store_true",
            default=False,
            help="Use batchnorm in mlp",
        )
        parser.add_argument(
            "--mlp_use_layer_norm",
            action="store_true",
            default=False,
            help="Use LayerNorm in mlp",
        )


@register_object("graph_classifier", "model")
class GraphClassifier(Classifier):
    def __init__(self, args):
        super(GraphClassifier, self).__init__(args)
        cargs = copy.deepcopy(args)
        cargs.mlp_layer_configuration = args.graph_classifier_mlp_layer_configuration
        cargs.mlp_use_batch_norm = args.classifier_mlp_use_batch_norm
        cargs.mlp_use_layer_norm = args.classifier_mlp_use_layer_norm
        if self.args.use_rdkit_features:
            cargs.mlp_input_dim = (
                args.graph_classifier_hidden_dim + args.rdkit_features_dim
            )
        self.mlp = get_object("mlp_classifier", "model")(cargs)

    def forward(self, batch=None):
        output = {}
        graph_x = self.encoder(batch)[self.encoder_hidden_key]
        output["encoder_hidden"] = graph_x
        batch_size = output["encoder_hidden"].shape[0]
        if self.args.use_rdkit_features:
            features = batch["rdkit_features"].view(batch_size, -1)
            graph_x = torch.concat([graph_x, features], dim=-1)
        output["hidden"] = graph_x
        output.update(self.mlp({"x": graph_x.float()}))
        return output

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        super(GraphClassifier, GraphClassifier).add_args(parser)
        parser.add_argument(
            "--use_rdkit_features",
            action="store_true",
            default=False,
            help="whether using graph-level features from rdkit",
        )
        parser.add_argument(
            "--rdkit_features_dim",
            type=int,
            default=0,
            help="number of features",
        )
        parser.add_argument(
            "--graph_classifier_hidden_dim",
            type=int,
            default=None,
            help="dimension of hidden layer from graph encoder",
        )
        parser.add_argument(
            "--graph_classifier_mlp_layer_configuration",
            type=int,
            nargs="*",
            default=[],
            help="MLP layer dimensions",
        )



import rdkit
from rdkit import Chem
import argparse
from rdkit import RDLogger
import numpy as np
from tqdm import tqdm
from indigo import *

lg = RDLogger.logger()
lg.setLevel(4)

idxfunc = lambda a: a.GetAtomMapNum()
bond_types = [
    Chem.rdchem.BondType.SINGLE,
    Chem.rdchem.BondType.DOUBLE,
    Chem.rdchem.BondType.TRIPLE,
    Chem.rdchem.BondType.AROMATIC,
]
bond_types_as_double = {0.0: 0, 1.0: 1, 2.0: 2, 3.0: 3, 1.5: 4}


# Define a standardization procedure so we can evaluate based on...
# a) RDKit-sanitized equivalence, and
# b) MOLVS-sanitized equivalence
from molvs import Standardizer

standardizer = Standardizer()
standardizer.prefer_organic = True


def sanitize_smiles(smi, largest_fragment=False):
    mol = Chem.MolFromSmiles(smi)
    if mol is None:
        return smi
    try:
        mol = standardizer.standardize(mol)  # standardize functional group reps
        if largest_fragment:
            mol = standardizer.largest_fragment(
                mol
            )  # remove product counterions/salts/etc.
        mol = standardizer.uncharge(mol)  # neutralize, e.g., carboxylic acids
    except Exception:
        pass
    return Chem.MolToSmiles(mol)


def addmapbyINDIGO(rxnsmi):
    indigo = Indigo()
    smi = rxnsmi
    rxn = indigo.loadReaction(smi)
    rxn.automap("discard ignore_charges ignore_isotopes ignore_valence ignore_radicals")
    rxnsmi_map_indigo = rxn.smiles()
    rxnsmi_map_indigo = rxn.canonicalSmiles()
    return rxnsmi_map_indigo


"""
Functions below evaluate the quality of predictions from the rank_diff_wln model by applying the predicted
graph edits to the reactants, cleaning up the generated product, and comparing it to what was recorded
as the true (major) product of that reaction
"""

# Define some post-sanitization reaction cleaning scripts
# These are to align our graph edit representation of a reaction with the data for improved coverage
from rdkit.Chem import AllChem

clean_rxns_presani = [
    AllChem.ReactionFromSmarts(
        "[O:1]=[c:2][n;H0:3]>>[O:1]=[c:2][n;H1:3]"
    ),  # hydroxypyridine written with carbonyl, must invent H on nitrogen
]
clean_rxns_postsani = [
    AllChem.ReactionFromSmarts(
        "[n;H1;+0:1]:[n;H0;+1:2]>>[n;H0;+0:1]:[n;H0;+0:2]"
    ),  # two adjacent aromatic nitrogens should allow for H shift
    AllChem.ReactionFromSmarts(
        "[n;H1;+0:1]:[c:3]:[n;H0;+1:2]>>[n;H0;+0:1]:[*:3]:[n;H0;+0:2]"
    ),  # two aromatic nitrogens separated by one should allow for H shift
    AllChem.ReactionFromSmarts("[#7;H0;+:1]-[O;H1;+0:2]>>[#7;H0;+:1]-[O;H0;-:2]"),
    AllChem.ReactionFromSmarts(
        "[C;H0;+0:1](=[O;H0;+0:2])[O;H0;-1:3]>>[C;H0;+0:1](=[O;H0;+0:2])[O;H1;+0:3]"
    ),  # neutralize C(=O)[O-]
    AllChem.ReactionFromSmarts(
        "[I,Br,F;H1;D0;+0:1]>>[*;H0;-1:1]"
    ),  # turn neutral halogens into anions EXCEPT HCl
    AllChem.ReactionFromSmarts(
        "[N;H0;-1:1]([C:2])[C:3]>>[N;H1;+0:1]([*:2])[*:3]"
    ),  # inexplicable nitrogen anion in reactants gets fixed in prods
]
for clean_rxn in clean_rxns_presani + clean_rxns_postsani:
    if clean_rxn.Validate() != (0, 0):
        raise ValueError("Invalid cleaning reaction - check your SMARTS!")
BOND_TYPE = [
    0,
    Chem.rdchem.BondType.SINGLE,
    Chem.rdchem.BondType.DOUBLE,
    Chem.rdchem.BondType.TRIPLE,
    Chem.rdchem.BondType.AROMATIC,
]


def edit_mol(rmol, edits):
    new_mol = Chem.RWMol(rmol)

    # Keep track of aromatic nitrogens, might cause explicit hydrogen issues
    aromatic_nitrogen_idx = set()
    aromatic_carbonyl_adj_to_aromatic_nH = {}
    aromatic_carbondeg3_adj_to_aromatic_nH0 = {}
    for a in new_mol.GetAtoms():
        if a.GetIsAromatic() and a.GetSymbol() == "N":
            aromatic_nitrogen_idx.add(a.GetIdx())
            for nbr in a.GetNeighbors():
                if (
                    a.GetNumExplicitHs() == 1
                    and nbr.GetSymbol() == "C"
                    and nbr.GetIsAromatic()
                    and any(b.GetBondTypeAsDouble() == 2 for b in nbr.GetBonds())
                ):
                    aromatic_carbonyl_adj_to_aromatic_nH[nbr.GetIdx()] = a.GetIdx()
                elif (
                    a.GetNumExplicitHs() == 0
                    and nbr.GetSymbol() == "C"
                    and nbr.GetIsAromatic()
                    and len(nbr.GetBonds()) == 3
                ):
                    aromatic_carbondeg3_adj_to_aromatic_nH0[nbr.GetIdx()] = a.GetIdx()
        else:
            a.SetNumExplicitHs(0)
    new_mol.UpdatePropertyCache()

    amap = {}
    for atom in rmol.GetAtoms():
        amap[atom.GetIntProp("molAtomMapNumber")] = atom.GetIdx()

    # Apply the edits as predicted
    for x, y, t in edits:
        bond = new_mol.GetBondBetweenAtoms(amap[x], amap[y])
        a1 = new_mol.GetAtomWithIdx(amap[x])
        a2 = new_mol.GetAtomWithIdx(amap[y])
        if bond is not None:
            new_mol.RemoveBond(amap[x], amap[y])

            # Are we losing a bond on an aromatic nitrogen?
            if bond.GetBondTypeAsDouble() == 1.0:
                if amap[x] in aromatic_nitrogen_idx:
                    if a1.GetTotalNumHs() == 0:
                        a1.SetNumExplicitHs(1)
                    elif a1.GetFormalCharge() == 1:
                        a1.SetFormalCharge(0)
                elif amap[y] in aromatic_nitrogen_idx:
                    if a2.GetTotalNumHs() == 0:
                        a2.SetNumExplicitHs(1)
                    elif a2.GetFormalCharge() == 1:
                        a2.SetFormalCharge(0)

            # Are we losing a c=O bond on an aromatic ring? If so, remove H from adjacent nH if appropriate
            if bond.GetBondTypeAsDouble() == 2.0:
                if amap[x] in aromatic_carbonyl_adj_to_aromatic_nH:
                    new_mol.GetAtomWithIdx(
                        aromatic_carbonyl_adj_to_aromatic_nH[amap[x]]
                    ).SetNumExplicitHs(0)
                elif amap[y] in aromatic_carbonyl_adj_to_aromatic_nH:
                    new_mol.GetAtomWithIdx(
                        aromatic_carbonyl_adj_to_aromatic_nH[amap[y]]
                    ).SetNumExplicitHs(0)

        if t > 0:
            new_mol.AddBond(amap[x], amap[y], BOND_TYPE[t])

            # Special alkylation case?
            if t == 1:
                if amap[x] in aromatic_nitrogen_idx:
                    if a1.GetTotalNumHs() == 1:
                        a1.SetNumExplicitHs(0)
                    else:
                        a1.SetFormalCharge(1)
                elif amap[y] in aromatic_nitrogen_idx:
                    if a2.GetTotalNumHs() == 1:
                        a2.SetNumExplicitHs(0)
                    else:
                        a2.SetFormalCharge(1)

            # Are we getting a c=O bond on an aromatic ring? If so, add H to adjacent nH0 if appropriate
            if t == 2:
                if amap[x] in aromatic_carbondeg3_adj_to_aromatic_nH0:
                    new_mol.GetAtomWithIdx(
                        aromatic_carbondeg3_adj_to_aromatic_nH0[amap[x]]
                    ).SetNumExplicitHs(1)
                elif amap[y] in aromatic_carbondeg3_adj_to_aromatic_nH0:
                    new_mol.GetAtomWithIdx(
                        aromatic_carbondeg3_adj_to_aromatic_nH0[amap[y]]
                    ).SetNumExplicitHs(1)

    pred_mol = new_mol.GetMol()

    # Clear formal charges to make molecules valid
    # Note: because S and P (among others) can change valence, be more flexible
    for atom in pred_mol.GetAtoms():
        atom.ClearProp("molAtomMapNumber")
        if (
            atom.GetSymbol() == "N" and atom.GetFormalCharge() == 1
        ):  # exclude negatively-charged azide
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if bond_vals <= 3:
                atom.SetFormalCharge(0)
        elif (
            atom.GetSymbol() == "N" and atom.GetFormalCharge() == -1
        ):  # handle negatively-charged azide addition
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if bond_vals == 3 and any(
                [nbr.GetSymbol() == "N" for nbr in atom.GetNeighbors()]
            ):
                atom.SetFormalCharge(0)
        elif atom.GetSymbol() == "N":
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if (
                bond_vals == 4 and not atom.GetIsAromatic()
            ):  # and atom.IsInRingSize(5)):
                atom.SetFormalCharge(1)
        elif atom.GetSymbol() == "C" and atom.GetFormalCharge() != 0:
            atom.SetFormalCharge(0)
        elif atom.GetSymbol() == "O" and atom.GetFormalCharge() != 0:
            bond_vals = (
                sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
                + atom.GetNumExplicitHs()
            )
            if bond_vals == 2:
                atom.SetFormalCharge(0)
        elif atom.GetSymbol() in ["Cl", "Br", "I", "F"] and atom.GetFormalCharge() != 0:
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if bond_vals == 1:
                atom.SetFormalCharge(0)
        elif atom.GetSymbol() == "S" and atom.GetFormalCharge() != 0:
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if bond_vals in [2, 4, 6]:
                atom.SetFormalCharge(0)
        elif (
            atom.GetSymbol() == "P"
        ):  # quartenary phosphorous should be pos. charge with 0 H
            bond_vals = [bond.GetBondTypeAsDouble() for bond in atom.GetBonds()]
            if sum(bond_vals) == 4 and len(bond_vals) == 4:
                atom.SetFormalCharge(1)
                atom.SetNumExplicitHs(0)
            elif sum(bond_vals) == 3 and len(bond_vals) == 3:  # make sure neutral
                atom.SetFormalCharge(0)
        elif atom.GetSymbol() == "B":  # quartenary boron should be neg. charge with 0 H
            bond_vals = [bond.GetBondTypeAsDouble() for bond in atom.GetBonds()]
            if sum(bond_vals) == 4 and len(bond_vals) == 4:
                atom.SetFormalCharge(-1)
                atom.SetNumExplicitHs(0)
        elif atom.GetSymbol() in ["Mg", "Zn"]:
            bond_vals = [bond.GetBondTypeAsDouble() for bond in atom.GetBonds()]
            if sum(bond_vals) == 1 and len(bond_vals) == 1:
                atom.SetFormalCharge(1)
        elif atom.GetSymbol() == "Si":
            bond_vals = [bond.GetBondTypeAsDouble() for bond in atom.GetBonds()]
            if sum(bond_vals) == len(bond_vals):
                atom.SetNumExplicitHs(max(0, 4 - len(bond_vals)))

    # Bounce to/from SMILES to try to sanitize
    pred_smiles = Chem.MolToSmiles(pred_mol)
    pred_list = pred_smiles.split(".")
    pred_mols = [Chem.MolFromSmiles(pred_smiles) for pred_smiles in pred_list]

    for i, mol in enumerate(pred_mols):
        # Check if we failed/succeeded in previous step
        if mol is None:
            # print('##### Unparseable mol: {}'.format(pred_list[i]))
            continue

        # Else, try post-sanitiztion fixes in structure
        mol = Chem.MolFromSmiles(Chem.MolToSmiles(mol))
        if mol is None:
            continue
        for rxn in clean_rxns_postsani:
            out = rxn.RunReactants((mol,))
            if out:
                try:
                    Chem.SanitizeMol(out[0][0])
                    pred_mols[i] = Chem.MolFromSmiles(Chem.MolToSmiles(out[0][0]))
                except Exception as e:
                    pass
                    # print(e)
                    # print('Could not sanitize postsani reaction product: {}'.format(Chem.MolToSmiles(out[0][0])))
                    # print('Original molecule was: {}'.format(Chem.MolToSmiles(mol)))
    pred_smiles = [
        Chem.MolToSmiles(pred_mol) for pred_mol in pred_mols if pred_mol is not None
    ]

    return pred_smiles


"""
Functions below the data used in Wengong Jin's NIPS paper on predicting reaction outcomes for the modified
forward prediction script. Rather than just training to predict which bonds change, we make a direct prediction
on HOW those bonds change
"""


def atom_numberize_post(reactants_smi):
    m = Chem.MolFromSmiles(reactants_smi)
    atomnums = sorted(
        [
            str(a.GetProp("molAtomMapNumber"))
            for a in m.GetAtoms()
            if a.HasProp("molAtomMapNumber")
        ]
    )

    mapnum = 1
    for a in m.GetAtoms():
        if not a.HasProp("molAtomMapNumber"):
            while str(mapnum) in atomnums:
                mapnum += 1
            a.SetIntProp("molAtomMapNumber", mapnum)
            mapnum += 1
        else:
            continue

    return Chem.MolToSmiles(m)


def get_changed_bonds(rxn_smi):
    reactants = Chem.MolFromSmiles(atom_numberize_post(rxn_smi.split(">")[0]))
    products = Chem.MolFromSmiles(rxn_smi.split(">")[2])

    conserved_maps = [
        a.GetProp("molAtomMapNumber")
        for a in products.GetAtoms()
        if a.HasProp("molAtomMapNumber")
    ]
    bond_changes = set()  # keep track of bond changes

    # Look at changed bonds
    bonds_prev = {}
    for bond in reactants.GetBonds():
        begin_atom_num = (
            bond.GetBeginAtom().GetProp("molAtomMapNumber")
            if bond.GetBeginAtom().HasProp("molAtomMapNumber")
            else None
        )
        end_atom_num = (
            bond.GetEndAtom().GetProp("molAtomMapNumber")
            if bond.GetEndAtom().HasProp("molAtomMapNumber")
            else None
        )
        nums = sorted([begin_atom_num, end_atom_num])

        if (nums[0] not in conserved_maps) and (nums[1] not in conserved_maps):
            continue
        bonds_prev["{}~{}".format(nums[0], nums[1])] = bond.GetBondTypeAsDouble()
    bonds_new = {}
    for bond in products.GetBonds():
        nums = sorted(
            [
                bond.GetBeginAtom().GetProp("molAtomMapNumber"),
                bond.GetEndAtom().GetProp("molAtomMapNumber"),
            ]
        )
        bonds_new["{}~{}".format(nums[0], nums[1])] = bond.GetBondTypeAsDouble()

    for bond in bonds_prev:
        if bond not in bonds_new:
            bond_changes.add((bond.split("~")[0], bond.split("~")[1], 0.0))  # lost bond
        else:
            if bonds_prev[bond] != bonds_new[bond]:
                bond_changes.add(
                    (bond.split("~")[0], bond.split("~")[1], bonds_new[bond])
                )  # changed bond
    for bond in bonds_new:
        if bond not in bonds_prev:
            bond_changes.add(
                (bond.split("~")[0], bond.split("~")[1], bonds_new[bond])
            )  # new bond

    return reactants, products, bond_changes


def product_is_recoverable(rmol, pmol, gedits, bonds_as_doubles=True):
    try:
        gfound, gfound_sani = 0, 0
        thisrow = []

        r = Chem.MolToSmiles(rmol)
        p = Chem.MolToSmiles(pmol)

        thisrow.append(r)
        thisrow.append(p)

        # Save pbond information
        pbonds = {}
        for bond in pmol.GetBonds():
            a1 = idxfunc(bond.GetBeginAtom())
            a2 = idxfunc(bond.GetEndAtom())
            t = bond_types.index(bond.GetBondType())
            pbonds[(a1, a2)] = pbonds[(a2, a1)] = t + 1

        for atom in pmol.GetAtoms():
            atom.ClearProp("molAtomMapNumber")

        psmiles = Chem.MolToSmiles(pmol)
        psmiles_sani = set(sanitize_smiles(psmiles, True).split("."))
        psmiles = set(psmiles.split("."))

        thisrow.append(".".join(psmiles))
        thisrow.append(".".join(psmiles_sani))

        ########### Use *true* edits to try to recover product

        if bonds_as_doubles:
            cbonds = []
            for gedit in gedits.split(";"):
                x, y, t = gedit.split("-")
                x, y, t = int(x), int(y), float(t)
                cbonds.append((x, y, bond_types_as_double[t]))
        else:
            # check if psmiles is recoverable
            cbonds = []
            for gedit in gedits.split(";"):
                x, y = gedit.split("-")
                x, y = int(x), int(y)
                if (x, y) in pbonds:
                    t = pbonds[(x, y)]
                else:
                    t = 0
                cbonds.append((x, y, t))

        # Generate products by modifying reactants with predicted edits.
        pred_smiles = edit_mol(rmol, cbonds)
        pred_smiles_sani = set(sanitize_smiles(smi) for smi in pred_smiles)
        pred_smiles = set(pred_smiles)
        if not psmiles <= pred_smiles:
            # Try again with kekulized form
            Chem.Kekulize(rmol)
            pred_smiles_kek = edit_mol(rmol, cbonds)
            pred_smiles_kek = set(pred_smiles_kek)
            if not psmiles <= pred_smiles_kek:
                if psmiles_sani <= pred_smiles_sani:
                    gfound_sani += 1

                else:
                    pass

            else:
                gfound += 1
                gfound_sani += 1

        else:
            gfound += 1
            gfound_sani += 1

        return gfound or gfound_sani
    except:
        return False


def get_atom_mapped_reaction(rxn_smi, args):
    try:
        rxn_smi = addmapbyINDIGO(rxn_smi)
    except:
        return

    rxn_smi = rxn_smi.strip().split(" ")[0]
    try:
        reactants, products, bond_changes = get_changed_bonds(rxn_smi)
        gedits = ";".join(["{}-{}-{}".format(x[0], x[1], x[2]) for x in bond_changes])
        is_recoverable = product_is_recoverable(
            reactants, products, gedits, bonds_as_doubles=not args.bonds_not_as_doubles
        )
        if not is_recoverable:
            return

    except:
        return

    rxn_smi = "{}>>{}".format(Chem.MolToSmiles(reactants), Chem.MolToSmiles(products))
    return {
        "reactants": Chem.MolToSmiles(reactants),
        "products": Chem.MolToSmiles(products),
        "bond_changes": gedits,
    }



import torch
from torch import optim
from clipzyme.utils.registry import register_object
from clipzyme.utils.classes import Nox


@register_object("sgd", "optimizer")
class SGD(optim.SGD, Nox):
    """
    https://pytorch.org/docs/stable/generated/torch.optim.SGD.html#torch.optim.SGD
    """

    def __init__(self, params, args):
        super().__init__(
            params=params,
            lr=args.lr,
            momentum=args.momentum,
            weight_decay=args.weight_decay,
        )


@register_object("adam", "optimizer")
class Adam(optim.Adam, Nox):
    """
    https://pytorch.org/docs/stable/generated/torch.optim.Adam.html#torch.optim.Adam
    """

    def __init__(self, params, args):
        adam_betas = tuple(args.adam_betas)
        super().__init__(
            params=params,
            lr=args.lr,
            betas=adam_betas,
            weight_decay=args.weight_decay,
            amsgrad=args.use_amsgrad,
        )

    @staticmethod
    def add_args(parser) -> None:
        parser.add_argument(
            "--use_amsgrad",
            action="store_true",
            default=False,
            help="whether to use the AMSGrad variant of Adam from the paper On the Convergence of Adam and Beyond",
        )
        parser.add_argument(
            "--adam_betas",
            type=float,
            nargs=2,
            default=[0.9, 0.999],
            help="momentume values for Adam optimizer",
        )


@register_object("adamw", "optimizer")
class AdamW(optim.AdamW, Nox):
    """
    https://pytorch.org/docs/stable/generated/torch.optim.AdamW.html#torch.optim.AdamW
    """

    def __init__(self, params, args):
        adam_betas = tuple(args.adam_betas)
        super().__init__(
            params=params,
            lr=args.lr,
            betas=adam_betas,
            weight_decay=args.weight_decay,
            amsgrad=args.use_amsgrad,
        )

    @staticmethod
    def add_args(parser) -> None:
        parser.add_argument(
            "--use_amsgrad",
            action="store_true",
            default=False,
            help="whether to use the AMSGrad variant of Adam from the paper On the Convergence of Adam and Beyond",
        )
        parser.add_argument(
            "--adam_betas",
            type=float,
            nargs=2,
            default=[0.9, 0.999],
            help="momentume values for Adam optimizer",
        )



from typing import List
import traceback, warnings, os, pickle
import argparse
import copy
import torch
from tqdm import tqdm
import Bio
import Bio.PDB
from Bio.Data.IUPACData import protein_letters_3to1
from esm import pretrained
from clipzyme.utils.registry import register_object
from clipzyme.utils.messages import METAFILE_NOTFOUND_ERR, LOAD_FAIL_MSG
from clipzyme.datasets.abstract import AbstractDataset
from clipzyme.utils.protein_utils import (
    read_structure_file,
    filter_resolution,
    build_graph,
    compute_graph_edges,
    compute_node_embedding,
)

protein_letters_3to1.update({k.upper(): v for k, v in protein_letters_3to1.items()})


@register_object("screening_enzymes", "dataset")
class ScreeningEnzymes(AbstractDataset):
    def __init__(self, args, split_group) -> None:
        if args.use_protein_graphs:
            self.esm_dir = args.esm_dir
            model, alphabet = pretrained.load_model_and_alphabet(args.esm_dir)
            self.esm_model = model
            self.alphabet = alphabet
            self.batch_converter = alphabet.get_batch_converter()
        super(ScreeningEnzymes, ScreeningEnzymes).__init__(self, args, split_group)

    def create_protein_graph(self, sample):
        try:
            raw_path = os.path.join(
                self.args.protein_structures_dir,
                f"AF-{sample['uniprot_id']}-F1-model_v4.cif",
            )
            protein_args = {
                "sample_id": sample["sample_id"],
                "protein_parser": Bio.PDB.MMCIFParser(),
                "protein_resolution": "residue",
                "graph_edge_args": {"knn_size": 10},
                "center_protein": True,
            }

            sample_id = protein_args["sample_id"]
            protein_parser = protein_args["protein_parser"]
            protein_resolution = protein_args["protein_resolution"]
            graph_edge_args = protein_args["graph_edge_args"]
            center_protein = protein_args["center_protein"]

            # parse pdb
            all_res, all_atom, all_pos = read_structure_file(
                protein_parser, raw_path, sample_id
            )
            # filter resolution of protein (backbone, atomic, etc.)
            atom_names, seq, pos = filter_resolution(
                all_res,
                all_atom,
                all_pos,
                protein_resolution=protein_resolution,
            )
            # generate graph
            data = build_graph(atom_names, seq, pos, sample_id)
            # kNN graph
            data = compute_graph_edges(data, **graph_edge_args)
            if center_protein:
                center = data["receptor"].pos.mean(dim=0, keepdim=True)
                data["receptor"].pos = data["receptor"].pos - center
                data.center = center
            uniprot_id = sample["uniprot_id"]
            sequence = self.uniprot2sequence[uniprot_id]
            data.structure_sequence = self.uniprot2sequence[uniprot_id]

            node_embeddings_args = {
                "model": self.esm_model,
                "model_location": self.esm_dir,
                "alphabet": self.alphabet,
                "batch_converter": self.batch_converter,
            }

            embedding_path = os.path.join(
                self.args.protein_graphs_dir,
                "precomputed_node_embeddings",
                f"{sample['uniprot_id']}.pt",
            )

            try:
                node_embedding = torch.load(sample["embedding_path"])
            except:
                node_embedding = compute_node_embedding(data, **node_embeddings_args)
            # Fix sequence length mismatches
            if len(data["receptor"].seq) != node_embedding.shape[0]:
                print("Computing seq embedding for mismatched seq length")
                protein_letters_3to1.update(
                    {k.upper(): v for k, v in protein_letters_3to1.items()}
                )
                AA_seq = ""
                for char in seq:
                    AA_seq += protein_letters_3to1[char]

                data.structure_sequence = AA_seq
                data["receptor"].x = compute_node_embedding(
                    data, **node_embeddings_args
                )
            else:
                data["receptor"].x = node_embedding

            if len(data["receptor"].seq) != data["receptor"].x.shape[0]:
                return None

            return data

        except Exception as e:
            print(
                f"Create prot graph: Could not load sample {sample['uniprot_id']} because of the exception {e}"
            )
            return None

    def load_dataset(self, args: argparse.ArgumentParser) -> None:
        """Loads dataset file

        Args:
            args (argparse.ArgumentParser)

        Raises:
            Exception: Unable to load
        """
        try:
            self.metadata_json = pickle.load(open(args.dataset_file_path, "rb"))
            self.alphafold_files = pickle.load(
                open("/home/datasets/alphafold_enzymes.p", "rb")
            )
            self.quickprot_caches = pickle.load(
                open("/home/datasets/quickprot_caches.p", "rb")
            )
            self.msa_files = pickle.load(
                open(
                    "/home/datasets/uniprot2msa_embedding.p",
                    "rb",
                )
            )
            self.uniprot2sequence = self.metadata_json
        except Exception as e:
            raise Exception(METAFILE_NOTFOUND_ERR.format(args.dataset_file_path, e))

    def create_dataset(self, split_group) -> List[dict]:
        dataset = []

        for uniprot_id, sequence in tqdm(self.metadata_json.items(), ncols=100):
            sample = {
                "uniprot_id": uniprot_id,
                "sequence": sequence,
                "sample_id": uniprot_id,
            }

            if self.skip_sample(sample):
                continue

            dataset.append(sample)

        return dataset

    def skip_sample(self, sample) -> bool:
        """
        Return True if sample should be skipped and not included in data
        """
        if "sequence" in sample:
            if (sample["sequence"] is None) or (len(sample["sequence"]) == 0):
                return True

            if (self.args.max_protein_length is not None) and len(
                sample["sequence"]
            ) > self.args.max_protein_length:
                return True

        if self.args.use_protein_graphs:
            if sample["uniprot_id"] not in self.alphafold_files:
                return True

        if self.args.use_protein_msa:
            if sample["uniprot_id"] not in self.msa_files:
                return True

        return False

    def get_split_group_dataset(self, processed_dataset, split_group):
        return processed_dataset

    def assign_splits(self, metadata_json, split_probs, seed=0) -> None:
        return metadata_json

    def __getitem__(self, index):
        """
        Fetch single sample from dataset

        Args:
            index (int): random index of sample from dataset

        Returns:
            sample (dict): a sample
        """
        sample = self.dataset[index]
        try:
            item = copy.deepcopy(sample)
            uniprot_id = item["uniprot_id"]

            if self.args.use_protein_graphs:
                if self.args.cache_path:
                    try:
                        graph_path_cache = os.path.join(
                            self.args.cache_path,
                            f"{item['uniprot_id']}_graph.pt",
                        )
                        data = torch.load(graph_path_cache)
                        if data is None:
                            data = self.load_protein_graph(item)
                            torch.save(data, graph_path_cache)
                    except:
                        data = self.load_protein_graph(item)
                        torch.save(data, graph_path_cache)
                else:
                    data = self.load_protein_graph(item)

                if self.args.use_protein_msa:
                    feats = data["receptor"].x
                    msa_embed = torch.load(self.msa_files[uniprot_id])
                    if self.args.replace_esm_with_msa:
                        data["receptor"].x = msa_embed
                    else:
                        data["receptor"].x = torch.concat([feats, msa_embed], dim=-1)
                        data["receptor"].msa = msa_embed

                item["graph"] = data

            return item
        except Exception:
            warnings.warn(
                LOAD_FAIL_MSG.format(sample["sample_id"], traceback.print_exc())
            )

    def load_protein_graph(self, item):
        # load the protein graph
        graph_path = os.path.join(
            self.args.protein_graphs_dir,
            "processed",
            f"{item['uniprot_id']}_graph.pt",
        )
        try:
            data = torch.load(graph_path)
        except:
            data = self.create_protein_graph(item)
            torch.save(data, graph_path)
        if data is None:
            try:
                data = self.create_protein_graph(item)
                torch.save(data, graph_path)
            except:
                return

        if hasattr(data, "x") and not hasattr(data["receptor"], "x"):
            data["receptor"].x = data.x

        if not hasattr(data, "structure_sequence"):
            data.structure_sequence = "".join(
                [protein_letters_3to1[char] for char in data["receptor"].seq]
            )

        keep_keys = {
            "receptor",
            "structure_sequence",
            ("receptor", "contact", "receptor"),
        }

        data_keys = data.to_dict().keys()
        for d_key in data_keys:
            if not d_key in keep_keys:
                delattr(data, d_key)

        coors = data["receptor"].pos
        feats = data["receptor"].x
        edge_index = data["receptor", "contact", "receptor"].edge_index
        assert (
            coors.shape[0] == feats.shape[0]
        ), f"Number of nodes do not match between coors ({coors.shape[0]}) and feats ({feats.shape[0]})"

        assert (
            max(edge_index[0]) < coors.shape[0] and max(edge_index[1]) < coors.shape[0]
        ), "Edge index contains node indices not present in coors"

        return data

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        AbstractDataset.add_args(parser)
        parser.add_argument(
            "--esm_dir",
            type=str,
            default="/home/snapshots/metabolomics/esm2/checkpoints/esm2_t33_650M_UR50D.pt",
            help="directory to load esm model from",
        )
        parser.add_argument(
            "--use_protein_graphs",
            action="store_true",
            default=False,
            help="whether to use and generate protein graphs",
        )
        parser.add_argument(
            "--protein_graphs_dir",
            type=str,
            default=None,
            help="directory to load protein graphs from",
        )
        parser.add_argument(
            "--protein_structures_dir",
            type=str,
            default=None,
            help="directory to load protein graphs from",
        )
        parser.add_argument(
            "--use_protein_msa",
            action="store_true",
            default=False,
            help="whether to use and generate protein MSAs",
        )
        parser.add_argument(
            "--protein_msa_dir",
            type=str,
            default="/home/datasets/EnzymeMap/embed_msa_transformer",
            help="directory where msa transformer embeddings are stored.",
        )
        parser.add_argument(
            "--max_protein_length",
            type=int,
            default=None,
            help="skip proteins longer than max_protein_length",
        )
        parser.add_argument(
            "--replace_esm_with_msa",
            action="store_true",
            default=False,
            help="whether to use ONLY the protein MSAs",
        )

    @staticmethod
    def set_args(args) -> None:
        args.dataset_file_path = (
            "/home/datasets/uniprot2sequence_standard_set_structs.p"
        )

    @property
    def SUMMARY_STATEMENT(self) -> None:
        """
        Prints summary statement with dataset stats
        """
        num_proteins = len(set([s["sequence"] for s in self.dataset]))
        statement = f""" 
        * Number of samples: {len(self.dataset)}
        * Number of proteins: {num_proteins}
        """
        return statement



from clipzyme.utils.registry import register_object
import torch
import torch.nn.functional as F
import torch.nn as nn
from collections import OrderedDict
import pdb
from clipzyme.utils.classes import Nox


@register_object("cross_entropy", "loss")
class CrossEntropyLoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        logging_dict, predictions = OrderedDict(), OrderedDict()
        logit = model_output["logit"]

        if "y" in model_output:
            predictions["golds"] = model_output["y"]
        elif "y" in batch:
            predictions["golds"] = batch["y"]
        else:
            raise KeyError("predictions_dict ERROR: y not found")

        target = predictions["golds"]
        if args.precomputed_loss:
            loss = model_output["loss"]
        else:
            loss = (
                F.cross_entropy(
                    logit.view(-1, args.num_classes), target.view(-1).long()
                )
                * args.ce_loss_lambda
            )
        logging_dict["cross_entropy_loss"] = loss.detach()
        predictions["probs"] = F.softmax(logit, dim=-1).detach()
        predictions["preds"] = predictions["probs"].argmax(axis=-1)

        if not args.keep_preds_dim:
            predictions["golds"] = predictions["golds"].view(-1)
            predictions["preds"] = predictions["preds"].reshape(-1)
        return loss, logging_dict, predictions

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--ce_loss_lambda",
            type=float,
            default=1.0,
            help="Lambda to weigh the cross-entropy loss.",
        )
        parser.add_argument(
            "--precomputed_loss",
            action="store_true",
            default=False,
            help="whether loss is computed through model automatically, e.g., hugging face transformers",
        )
        parser.add_argument(
            "--keep_preds_dim",
            action="store_true",
            default=False,
            help="do not flatten preds and y",
        )


@register_object("binary_cross_entropy_logits", "loss")
class BinaryCrossEntropyLoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        logging_dict, predictions = OrderedDict(), OrderedDict()
        logit = model_output["logit"]
        if logit.shape[-1] == 1 and len(logit.shape) == 2:
            logit = logit.squeeze(-1)

        if "has_y" in batch:
            loss = (
                F.binary_cross_entropy_with_logits(
                    logit, batch["y"].float(), reduction="none", weight=batch["has_y"]
                ).sum()
                / batch["has_y"].sum()
                * args.ce_loss_lambda
            )
            predictions["has_golds"] = batch["has_y"]
        else:
            loss = (
                F.binary_cross_entropy_with_logits(logit, batch["y"].float())
                * args.ce_loss_lambda
            )
        logging_dict["binary_cross_entropy_loss"] = loss.detach()
        if args.use_top_prediction:
            probs = torch.sigmoid(logit).detach()
            predictions["probs"] = torch.max(probs, dim=-1)
            prob_ids = torch.argmax(probs, dim=-1)
            predictions["golds"] = batch["y"][prob_ids]
            predictions["preds"] = predictions["probs"] > 0.5
        else:
            predictions["probs"] = torch.sigmoid(logit).detach()
            predictions["golds"] = batch["y"]
            predictions["preds"] = predictions["probs"] > 0.5
        return loss, logging_dict, predictions

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--ce_loss_lambda",
            type=float,
            default=1.0,
            help="Lambda to weigh the cross-entropy loss.",
        )
        parser.add_argument(
            "--use_top_prediction",
            action="store_true",
            default=False,
            help="store top prediction in predictions dict",
        )


@register_object("mse", "loss")
class MSELoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        l_dict, p_dict = OrderedDict(), OrderedDict()
        B = batch["y"].shape[0]
        golds = batch["y"].view(B, -1).float()
        logit = model_output["logit"].view(B, -1)
        loss = F.mse_loss(logit, golds) * args.mse_loss_lambda
        l_dict["mse_loss"] = loss.detach()
        p_dict["probs"] = logit.detach()
        p_dict["golds"] = golds
        return loss, l_dict, p_dict

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args
        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--mse_loss_lambda",
            type=float,
            default=1.0,
            help="Lambda to weigh the MSE loss.",
        )


@register_object("rmse", "loss")
class RMSELoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        l_dict, p_dict = OrderedDict(), OrderedDict()
        B = batch["y"].shape[0]
        golds = batch["y"].view(B, -1).float()
        logit = model_output["logit"].view(B, -1)
        loss = (
            torch.sqrt(F.mse_loss(logit, golds) + args.rmse_eps) * args.rmse_loss_lambda
        )
        l_dict["rmse_loss"] = loss.detach()
        p_dict["probs"] = logit.detach()
        p_dict["golds"] = golds
        return loss, l_dict, p_dict

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args
        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--rmse_loss_lambda",
            type=float,
            default=1.0,
            help="Lambda to weigh the RMSE loss.",
        )
        parser.add_argument(
            "--rmse_eps",
            type=float,
            default=1e-8,
            help="epsilon to avoid nan when backpropagating",
        )


@register_object("smoothl1", "loss")
class SmoothL1Loss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        l_dict, p_dict = OrderedDict(), OrderedDict()
        B = batch["y"].shape[0]
        golds = batch["y"].view(B, -1).float()
        logit = model_output["logit"].view(B, -1)
        loss = (
            F.smooth_l1_loss(logit, golds, beta=args.smoothl1_beta)
            * args.smoothl1_loss_lambda
        )
        l_dict["smoothl1_loss"] = loss.detach()
        p_dict["probs"] = logit.detach()
        p_dict["golds"] = golds
        return loss, l_dict, p_dict

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args
        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--smoothl1_loss_lambda",
            type=float,
            default=1.0,
            help="Lambda to weigh the SmoothL1 loss.",
        )
        parser.add_argument(
            "--smoothl1_beta",
            type=float,
            default=1.0,
            help="Specifies the threshold at which to change between L1 and L2 loss. The value must be non-negative.",
        )


@register_object("mae", "loss")
class MAELoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        l_dict, p_dict = OrderedDict(), OrderedDict()
        B = batch["y"].shape[0]
        golds = batch["y"].view(B, -1).float()
        logit = model_output["logit"].view(B, -1)
        loss = F.l1_loss(logit, golds) * args.mae_loss_lambda
        l_dict["mae_loss"] = loss.detach()
        p_dict["probs"] = logit.detach()
        p_dict["golds"] = golds
        return loss, l_dict, p_dict

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args
        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--mae_loss_lambda",
            type=float,
            default=1.0,
            help="Lambda to weigh the MSE loss.",
        )


@register_object("survival", "loss")
class SurvivalLoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        logging_dict, predictions = OrderedDict(), OrderedDict()
        logit = model_output["logit"]
        y_seq, y_mask = batch["y_seq"], batch["y_mask"]
        loss = F.binary_cross_entropy_with_logits(
            logit, y_seq.float(), weight=y_mask.float(), reduction="sum"
        ) / torch.sum(y_mask.float())
        logging_dict["survival_loss"] = loss.detach()
        predictions["probs"] = torch.sigmoid(logit).detach()
        predictions["golds"] = batch["y"]
        predictions["censors"] = batch["time_at_event"]
        return loss, logging_dict, predictions


@register_object("ordinal_cross_entropy", "loss")
class RankConsistentLoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        """
        Computes cross-entropy loss

        If batch contains they key 'has_y', the cross entropy loss will be computed for samples where batch['has_y'] = 1
        Expects model_output to contain 'logit'

        Returns:
            loss: cross entropy loss
            l_dict (dict): dictionary containing cross_entropy_loss detached from computation graph
            p_dict (dict): dictionary of model predictions and ground truth labels (preds, probs, golds)
        """
        loss = 0
        l_dict, p_dict = OrderedDict(), OrderedDict()
        logit = model_output["logit"]
        yseq = batch["yseq"]
        ymask = batch["ymask"]

        loss = F.binary_cross_entropy_with_logits(
            logit, yseq.float(), weight=ymask.float(), reduction="sum"
        ) / torch.sum(ymask.float())

        probs = F.logsigmoid(logit)  # log_sum to add probs
        probs = probs.unsqueeze(1).repeat(1, len(args.rank_thresholds), 1)
        probs = torch.tril(probs).sum(2)
        probs = torch.exp(probs)

        p_dict["logits"] = logit.detach()
        p_dict["probs"] = probs.detach()
        preds = probs > 0.5  # class = last prob > 0.5
        preds = preds.sum(-1)
        p_dict["preds"] = preds
        p_dict["golds"] = batch["y"]

        return loss, l_dict, p_dict



import argparse
import os
import pwd
from pytorch_lightning import Trainer
import itertools
from clipzyme.utils.registry import md5
import json
import copy
from clipzyme.utils.classes import set_nox_type
import inspect

EMPTY_NAME_ERR = 'Name of augmentation or one of its arguments cant be empty\n\
                  Use "name/arg1=value/arg2=value" format'
POSS_VAL_NOT_LIST = (
    "Flag {} has an invalid list of values: {}. Length of list must be >=1"
)


class GlobalNamespace(argparse.Namespace):
    pass


def parse_dispatcher_config(config):
    """
    Parses an experiment config, and creates jobs. For flags that are expected to be a single item,
    but the config contains a list, this will return one job for each item in the list.
    :config - experiment_config

    returns: jobs - a list of flag strings, each of which encapsulates one job.
         *Example: --train --cuda --dropout=0.1 ...
    returns: experiment_axies - axies that the grid search is searching over
    """

    assert all(
        [
            k
            in [
                "script",
                "available_gpus",
                "cartesian_hyperparams",
                "paired_hyperparams",
                "tune_hyperparams",
            ]
            for k in config.keys()
        ]
    )

    cartesian_hyperparamss = config["cartesian_hyperparams"]
    paired_hyperparams = config.get("paired_hyperparams", [])
    flags = []
    arguments = []
    experiment_axies = []

    # add anything outside search space as fixed
    fixed_args = ""
    for arg in config:
        if arg not in [
            "script",
            "cartesian_hyperparams",
            "paired_hyperparams",
            "available_gpus",
        ]:
            if type(config[arg]) is bool:
                if config[arg]:
                    fixed_args += "--{} ".format(str(arg))
                else:
                    continue
            else:
                fixed_args += "--{} {} ".format(arg, config[arg])

    # add paired combo of search space
    paired_args_list = [""]
    if len(paired_hyperparams) > 0:
        paired_args_list = []
        paired_keys = list(paired_hyperparams.keys())
        paired_vals = list(paired_hyperparams.values())
        flags.extend(paired_keys)
        for paired_combo in zip(*paired_vals):
            paired_args = ""
            for i, flg_value in enumerate(paired_combo):
                if type(flg_value) is bool:
                    if flg_value:
                        paired_args += "--{} ".format(str(paired_keys[i]))
                    else:
                        continue
                else:
                    paired_args += "--{} {} ".format(
                        str(paired_keys[i]), str(flg_value)
                    )
            paired_args_list.append(paired_args)

    # add every combo of search space
    product_flags = []
    for key, value in cartesian_hyperparamss.items():
        flags.append(key)
        product_flags.append(key)
        arguments.append(value)
        if len(value) > 1:
            experiment_axies.append(key)

    experiments = []
    exps_combs = list(itertools.product(*arguments))

    for tpl in exps_combs:
        exp = ""
        for idx, flg in enumerate(product_flags):
            if type(tpl[idx]) is bool:
                if tpl[idx]:
                    exp += "--{} ".format(str(flg))
                else:
                    continue
            else:
                exp += "--{} {} ".format(str(flg), str(tpl[idx]))
        exp += fixed_args
        for paired_args in paired_args_list:
            experiments.append(exp + paired_args)

    return experiments, flags, experiment_axies


def prepare_training_config_for_eval(train_config):
    """Convert training config to an eval config for testing.

    Parameters
    ----------
    train_config: dict
         config with the following structure:
              {
                   "train_config": ,   # path to train config
                   "log_dir": ,        # log directory used by dispatcher during training
                   "eval_args": {}     # test set-specific arguments beyond default
              }

    Returns
    -------
    experiments: list
    flags: list
    experiment_axies: list
    """

    train_args = json.load(open(train_config["train_config"], "r"))

    experiments, _, _ = parse_dispatcher_config(train_args)
    stem_names = [md5(e) for e in experiments]
    eval_args = copy.deepcopy(train_args)
    eval_args["cartesian_hyperparams"].update(train_config["eval_args"])

    # reset defaults
    eval_args["cartesian_hyperparams"]["train"] = [False]
    eval_args["cartesian_hyperparams"]["test"] = [True]
    eval_args["cartesian_hyperparams"]["from_checkpoint"] = train_config[
        "eval_args"
    ].get("from_checkpoint", [True])
    eval_args["cartesian_hyperparams"]["gpus"] = [1]
    eval_args["cartesian_hyperparams"]["logger_tags"][0] += " eval"
    eval_args["available_gpus"] = train_config["available_gpus"]
    eval_args["script"] = train_config["script"]

    experiments, flags, experiment_axies = parse_dispatcher_config(eval_args)

    if "checkpoint_path" not in eval_args["cartesian_hyperparams"]:
        for (idx, e), s in zip(enumerate(experiments), stem_names):
            experiments[idx] += " --checkpoint_path {}".format(
                os.path.join(train_config["log_dir"], "{}.args".format(s))
            )

    return experiments, flags, experiment_axies


def parse_augmentations(augmentations):
    """
    Parse the list of augmentations, given by configuration, into a list of
    tuple of the augmentations name and a dictionary containing additional args.

    The augmentation is assumed to be of the form 'name/arg1=value/arg2=value'

    :raw_augmentations: list of strings [unparsed augmentations]
    :returns: list of parsed augmentations [list of (name,additional_args)]

    """
    raw_transformers = augmentations

    transformers = []
    for t in raw_transformers:
        arguments = t.split("/")
        name = arguments[0]
        if name == "":
            raise Exception(EMPTY_NAME_ERR)

        kwargs = {}
        if len(arguments) > 1:
            for a in arguments[1:]:
                splited = a.split("=")
                var = splited[0]
                val = splited[1] if len(splited) > 1 else None
                if var == "":
                    raise Exception(EMPTY_NAME_ERR)
                try:
                    kwargs[var] = float(val)
                except ValueError:
                    kwargs[var] = val

        transformers.append((name, kwargs))

    return transformers


def get_parser():
    global_namespace = GlobalNamespace(allow_abbrev=False)

    parser = argparse.ArgumentParser(
        description="Nox Standard Args.", allow_abbrev=False
    )

    # -------------------------------------
    # Run Setup
    # -------------------------------------
    parser.add_argument(
        "--train",
        action="store_true",
        default=False,
        help="Whether or not to train model",
    )
    parser.add_argument(
        "--dev",
        action="store_true",
        default=False,
        help="Whether or not to run model on dev set",
    )
    parser.add_argument(
        "--test",
        action="store_true",
        default=False,
        help="Whether or not to run model on test set",
    )
    parser.add_argument(
        "--predict",
        action="store_true",
        default=False,
        help="Whether to run model for pure prediction where labels are not known",
    )
    parser.add_argument(
        "--eval_on_train",
        action="store_true",
        default=False,
        help="Whether or not to evaluate model on train split",
    )
    parser.add_argument(
        "--eval_on_train_multigpu",
        action="store_true",
        default=False,
        help="Whether or not to evaluate model on train split using ddp",
    )
    parser.add_argument(
        "--replicate",
        type=int,
        default=1,
        help="The replicate number for the experiment for running same experiments multiple times",
    )
    parser.add_argument(
        "--shuffle_eval_loader",
        action="store_true",
        default=False,
        help="shuffle the dev and test datasets",
    )

    # -------------------------------------
    # Data
    # -------------------------------------
    parser.add_argument(
        "--dataset_name",
        type=str,
        action=set_nox_type("dataset"),
        default="mnist",
        help="Name of dataset",
    )
    parser.add_argument(
        "--img_size",
        type=int,
        nargs="+",
        default=[256, 256],
        help="Width and height of image in pixels. [default: [256,256]]",
    )
    parser.add_argument(
        "--num_chan", type=int, default=3, help="Number of channels for input image"
    )
    parser.add_argument(
        "--img_mean",
        type=float,
        nargs="+",
        default=[128.1722],
        help="Mean of image per channel",
    )
    parser.add_argument(
        "--img_std",
        type=float,
        nargs="+",
        default=[87.1849],
        help="Standard deviation  of image per channel",
    )
    parser.add_argument(
        "--img_file_type",
        type=str,
        default="png",
        choices=["png", "dicom"],
        help="Type of image. one of [png, dicom]",
    )

    # -------------------------------------
    # Augmentations
    # -------------------------------------
    parser.add_argument(
        "--train_rawinput_augmentation_names",
        nargs="*",
        action=set_nox_type("augmentation"),
        default=[],
        help='List of image-transformations to use. Usage: "--train_rawinput_augmentations trans1/arg1=5/arg2=2 trans2 trans3/arg4=val"',
    )
    parser.add_argument(
        "--train_tnsr_augmentation_names",
        nargs="*",
        action=set_nox_type("augmentation"),
        default=[],
        help='List of image-transformations to use. Usage: "--train_tnsr_augmentations trans1/arg1=5/arg2=2 trans2 trans3/arg4=val"',
    )
    parser.add_argument(
        "--test_rawinput_augmentation_names",
        nargs="*",
        action=set_nox_type("augmentation"),
        default=[],
        help="List of image-transformations to use for the dev and test dataset",
    )
    parser.add_argument(
        "--test_tnsr_augmentation_names",
        nargs="*",
        action=set_nox_type("augmentation"),
        default=[],
        help="List of image-transformations to use for the dev and test dataset",
    )

    # -------------------------------------
    # Losses
    # -------------------------------------

    # losses and metrics
    parser.add_argument(
        "--loss_names",
        type=str,
        action=set_nox_type("loss"),
        nargs="*",
        default=[],
        help="Name of loss",
    )
    parser.add_argument(
        "--loss_names_for_eval",
        type=str,
        action=set_nox_type("loss"),
        nargs="*",
        default=None,
        help="Name of loss",
    )

    # -------------------------------------
    # Metrics
    # -------------------------------------

    parser.add_argument(
        "--metric_names",
        type=str,
        action=set_nox_type("metric"),
        nargs="*",
        default=[],
        help="Name of performance metric",
    )
    parser.add_argument(
        "--metric_names_for_eval",
        type=str,
        action=set_nox_type("metric"),
        nargs="*",
        default=None,
        help="Name of metric",
    )

    # -------------------------------------
    # Training Module
    # -------------------------------------

    parser.add_argument(
        "--lightning_name",
        type=str,
        action=set_nox_type("lightning"),
        default="base",
        help="Name of lightning module",
    )

    # -------------------------------------
    # Hyper parameters
    # -------------------------------------
    # learning
    parser.add_argument(
        "--batch_size",
        type=int,
        default=32,
        help="Batch size for training [default: 128]",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=0.001,
        help="Initial learning rate [default: 0.001]",
    )
    parser.add_argument(
        "--dropout",
        type=float,
        default=0.25,
        help="Amount of dropout to apply on last hidden layer [default: 0.25]",
    )
    parser.add_argument(
        "--optimizer_name",
        type=str,
        action=set_nox_type("optimizer"),
        default="adam",
        help="Optimizer to use [default: adam]",
    )
    parser.add_argument(
        "--momentum", type=float, default=0, help="Momentum to use with SGD"
    )
    parser.add_argument(
        "--lr_decay",
        type=float,
        default=0.1,
        help="Initial learning rate [default: 0.5]",
    )
    parser.add_argument(
        "--weight_decay",
        type=float,
        default=0,
        help="L2 Regularization penaty [default: 0]",
    )

    # tune
    parser.add_argument(
        "--tune_hyperopt",
        action="store_true",
        default=False,
        help="Whether to run hyper-parameter optimization",
    )
    parser.add_argument(
        "--tune_search_alg",
        type=str,
        default="search",
        help="Optimization algorithm",
    )
    parser.add_argument(
        "--tune_hyperparam_names",
        type=str,
        nargs="*",
        default=[],
        help="Name of parameters being optimized",
    )

    # -------------------------------------
    # Schedule
    # -------------------------------------
    parser.add_argument(
        "--scheduler_name",
        type=str,
        action=set_nox_type("scheduler"),
        default="reduce_on_plateau",
        help="Name of scheduler",
    )
    parser.add_argument(
        "--cosine_annealing_period",
        type=int,
        default=10,
        help="length of period of lr cosine anneal",
    )
    parser.add_argument(
        "--cosine_annealing_period_scaling",
        type=int,
        default=2,
        help="how much to multiply each period in successive annealing",
    )
    parser.add_argument(
        "--patience",
        type=int,
        default=5,
        help="Number of epochs without improvement on dev before halving learning rate and reloading best model [default: 5]",
    )
    parser.add_argument(
        "--num_adv_steps",
        type=int,
        default=1,
        help="Number of steps for domain adaptation discriminator per one step of encoding model [default: 5]",
    )

    # -------------------------------------
    # Callbacks
    # -------------------------------------

    parser.add_argument(
        "--callback_names",
        type=str,
        action=set_nox_type("callback"),
        nargs="*",
        default=["checkpointer", "lr_monitor"],
        help="Lightning callbacks",
    )

    parser.add_argument(
        "--monitor",
        type=str,
        default=None,
        help="Name of metric to use to decide when to save model",
    )

    parser.add_argument(
        "--checkpoint_save_top_k",
        type=int,
        default=1,
        help="the best k models according to the quantity monitored will be saved",
    )
    parser.add_argument(
        "--checkpoint_save_last",
        action="store_true",
        default=False,
        help="save the last model to last.ckpt",
    )

    # -------------------------------------
    # Model checkpointing
    # -------------------------------------

    parser.add_argument(
        "--checkpoint_dir", type=str, default="snapshot", help="Where to dump the model"
    )
    parser.add_argument(
        "--from_checkpoint",
        action="store_true",
        default=False,
        help="Whether loading a model from a saved checkpoint",
    )
    parser.add_argument(
        "--relax_checkpoint_matching",
        action="store_true",
        default=False,
        help="Do not enforce that the keys in checkpoint_path match the keys returned by this module’s state dict",
    )
    parser.add_argument(
        "--checkpoint_path",
        type=str,
        default=None,
        help="Filename of model snapshot to load[default: None]",
    )

    # -------------------------------------
    # Storing model outputs
    # -------------------------------------
    parser.add_argument(
        "--save_hiddens",
        action="store_true",
        default=False,
        help="Save hidden repr from each image to an npz based off results path, git hash and exam name",
    )
    parser.add_argument(
        "--save_predictions",
        action="store_true",
        default=False,
        help="Save hidden repr from each image to an npz based off results path, git hash and exam name",
    )
    parser.add_argument(
        "--inference_dir",
        type=str,
        default="hiddens/test_run",
        help='Dir to store hiddens npy"s when store_hiddens is true',
    )

    # -------------------------------------
    # Run outputs
    # -------------------------------------
    parser.add_argument(
        "--results_path",
        type=str,
        default="logs/test.args",
        help="Where to save the result logs",
    )
    parser.add_argument(
        "--experiment_name",
        type=str,
        help="defined either automatically by dispatcher.py or time in main.py. Keep without default",
    )

    # -------------------------------------
    # System
    # -------------------------------------
    parser.add_argument(
        "--num_workers",
        type=int,
        default=8,
        help="Num workers for each data loader [default: 4]",
    )

    # cache
    parser.add_argument(
        "--cache_path", type=str, default=None, help="Dir to cache images."
    )

    # -------------------------------------
    # Logging
    # -------------------------------------

    parser.add_argument(
        "--logger_name",
        type=str,
        action=set_nox_type("logger"),
        choices=["tensorboard", "comet", "wandb"],
        default="tensorboard",
        help="experiment logger to use",
    )
    parser.add_argument(
        "--logger_tags", nargs="*", default=[], help="List of tags for logger"
    )
    parser.add_argument("--project_name", default="CancerCures", help="Comet project")
    parser.add_argument("--workspace", default=None, help="Comet workspace")
    parser.add_argument(
        "--log_gen_image",
        action="store_true",
        default=False,
        help="Whether to log sample generated image to comet",
    )
    parser.add_argument(
        "--log_profiler",
        action="store_true",
        default=False,
        help="Log profiler times to logger",
    )

    # -------------------------------------
    # Add object-level args
    # -------------------------------------

    def add_class_args(args_as_dict, parser):
        # for loop
        for argname, argval in args_as_dict.items():
            args_for_noxs = {
                a.dest: a for a in parser._actions if hasattr(a, "is_nox_action")
            }
            old_args = vars(parser.parse_known_args()[0])
            if argname in args_for_noxs:
                args_for_noxs[argname].add_args(parser, argval)
                newargs = vars(parser.parse_known_args()[0])
                newargs = {k: v for k, v in newargs.items() if k not in old_args}
                add_class_args(newargs, parser)

    parser.parse_known_args(namespace=global_namespace)
    add_class_args(vars(global_namespace), parser)

    return parser


def parse_args(args_strings=None):
    # run
    # Lightning 2.0 removes add_argparse_args
    parser = get_parser()
    sig = inspect.signature(Trainer.__init__)
    for name, param in sig.parameters.items():
        if name == "self":
            continue
        if param.default is inspect.Parameter.empty:
            parser.add_argument(f"--{name}")
        else:
            # Note that this does not store types, because many defaults are None
            parser.add_argument(f"--{name}", default=param.default)

    # legacy
    if not hasattr(Trainer, "add_argparse_args"):
        parser.add_argument(
            "--gpus",
            default=None,
            help="Number of GPUs to train on",
        )

    if args_strings is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args_strings)

    # legacy - fix arg changes
    if hasattr(args, "gpus") and not hasattr(Trainer, "add_argparse_args"):
        args.devices = args.gpus

    # using gpus
    if (isinstance(args.gpus, str) and len(args.gpus.split(",")) > 1) or (
        isinstance(args.gpus, int) and args.gpus > 1
    ):
        args.strategy = "ddp"
        # should not matter since we set our sampler, in 2.0 this is default True and used to be called replace_sampler_ddp
        args.use_distributed_sampler = False
    else:
        if not hasattr(Trainer, "add_argparse_args"):  # lightning 2.0
            # args.strategy = "auto"
            args.strategy = "ddp"  # should be overwritten later in main
        else:  # legacy
            args.strategy = None
            args.replace_sampler_ddp = False

    # username
    args.unix_username = pwd.getpwuid(os.getuid())[0]

    # learning initial state
    args.step_indx = 1

    # set args
    args_for_noxs = {a.dest: a for a in parser._actions if hasattr(a, "is_nox_action")}
    for argname, argval in vars(args).items():
        if argname in args_for_noxs:
            args_for_noxs[argname].set_args(args, argval)

    # parse augmentations
    args.train_rawinput_augmentations = parse_augmentations(
        args.train_rawinput_augmentation_names
    )
    args.train_tnsr_augmentations = parse_augmentations(
        args.train_tnsr_augmentation_names
    )
    args.test_rawinput_augmentations = parse_augmentations(
        args.test_rawinput_augmentation_names
    )
    args.test_tnsr_augmentations = parse_augmentations(
        args.test_tnsr_augmentation_names
    )

    # parse tune parameters
    # args = parse_tune_params(args)

    return args



from clipzyme.utils.registry import register_object
import torch
import torch.nn.functional as F
import torch.nn as nn
from collections import OrderedDict
from clipzyme.utils.classes import Nox
from clipzyme.utils.loading import concat_all_gather
import torch.distributed as dist

EPSILON = 1e-6


def off_diagonal(x):
    n, m = x.shape
    assert n == m
    return x.flatten()[:-1].view(n - 1, n + 1)[:, 1:].flatten()


@register_object("ntxent_loss", "loss")
class NTexntLoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        """
        Computes contrastive loss batch-wise

        Expects model_output to contain [projection_hidden_1, projection_hidden2, prediction_hidden_1, prediction_hidden_2]

        Returns:
            loss: cross contrastive loss with stop grad
            l_dict (dict): dictionary containing simsiam_loss detached from computation graph
            p_dict (dict): dictionary of model's reconstruction of x
        """
        l_dict, p_dict = OrderedDict(), OrderedDict()

        N = model_output["reaction_projection"].shape[0] - 1
        p1, p2 = (
            model_output["reaction_projection"],
            model_output["sequence_projection"],
        )

        # Normalize
        p1 = nn.functional.normalize(p1, dim=1)
        p2 = nn.functional.normalize(p2, dim=1)

        # calculate negative cosine similarity
        similarity = torch.einsum("ic, jc -> ij", p1, p2)

        if args.use_hard_sampling:
            neg_terms = (similarity.fill_diagonal(0),)
            pos_terms = torch.diag(similarity)
            reweight = (args.beta * neg_terms) / neg_terms.mean()
            Neg = max(
                (-N * args.tau_plus * pos_terms + reweight * neg_terms).sum()
                / (1 - args.tau_plus),
                torch.exp(-1 / args.temperature),
            )
            loss = -torch.log(pos_terms.sum() / (pos_terms.sum() + Neg))
        else:
            y = torch.arange(similarity.shape[0], device=similarity.device)
            loss = F.cross_entropy(similarity, y)
            l_dict["contrastive_loss"] = loss.detach()

        logits = similarity.detach()
        probs = torch.softmax(logits, dim=-1)
        if logits.shape[-1] != args.batch_size:
            logits = torch.cat(
                [
                    logits,
                    torch.zeros(
                        logits.shape[0],
                        args.batch_size - logits.shape[1],
                        device=similarity.device,
                    ),
                ],
                dim=-1,
            )
            probs = torch.cat(
                [
                    probs,
                    torch.zeros(
                        probs.shape[0],
                        args.batch_size - probs.shape[1],
                        device=similarity.device,
                    ),
                ],
                dim=-1,
            )
        p_dict["logits"] = logits
        p_dict["probs"] = probs
        p_dict["golds"] = y

        p_dict["reaction_hidden"] = model_output["reaction_hidden"].detach()
        p_dict["sequence_hidden"] = model_output["sequence_hidden"].detach()

        return loss, l_dict, p_dict

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--ntxent_loss_lambda",
            type=float,
            default=1.0,
            help="Lambda to weigh the cross-entropy loss.",
        )
        parser.add_argument(
            "--use_hard_sampling",
            action="store_true",
            default=False,
            help="Whether to use hard negative sampling.",
        )
        parser.add_argument(
            "--tau_plus",
            type=float,
            default=1.0,
            help="prior over probability of positive class.",
        )
        parser.add_argument(
            "--beta",
            type=float,
            default=1.0,
            help="hardness value.",
        )
        parser.add_argument(
            "--temperature",
            type=float,
            default=1.0,
            help="temperature value.",
        )


@register_object("barlow_loss", "loss")
class BarlowLoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        logging_dict, predictions = OrderedDict(), OrderedDict()

        z1 = model_output["reaction_projection"]
        z2 = model_output["sequence_projection"]

        # normalize repr. along the batch dimension

        z1_norm = (z1 - torch.mean(z1, dim=0)) / (torch.std(z1, dim=0) + EPSILON)
        z2_norm = (z2 - torch.mean(z2, dim=0)) / (torch.std(z2, dim=0) + EPSILON)

        # cross-correlation matrix
        cor_mat = torch.mm(z1_norm.T, z2_norm) / z1.shape[0]
        # if args.distributed_backend == 'ddp':
        #    torch.distributed.all_reduce(cor_mat)

        # loss
        on_diag = torch.diagonal(cor_mat).add_(-1).pow_(2).sum()
        off_diag = off_diagonal(cor_mat).pow_(2).sum()
        loss = args.barlow_loss_lambda * (
            on_diag + args.redundancy_reduction_lambda * off_diag
        )

        logging_dict["barlow_loss"] = loss.detach()

        predictions["reaction_hidden"] = model_output["reaction_hidden"].detach()
        predictions["sequence_hidden"] = model_output["sequence_hidden"].detach()

        return loss, logging_dict, predictions

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--barlow_loss_lambda",
            type=float,
            default=1.0,
            help="Lambda to weigh the cross-entropy loss.",
        )


@register_object("dclw_loss", "loss")
class DCLWLoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        """
        Computes decoupled contrastive learning loss

        Expects model_output to contain [projection_hidden_1, projection_hidden2, prediction_hidden_1, prediction_hidden_2]

        Returns:
            loss: dclw loss
            l_dict (dict): dictionary containing barlow_loss detached from computation graph
            p_dict (dict): dictionary of model's reconstruction of x
        """

        def von_mises_fisher_w(sim_vec):
            n = torch.exp(sim_vec / args.dclw_sigma)
            return 2 - n / n.mean()

        l_dict, p_dict = OrderedDict(), OrderedDict()

        z1, z2 = (
            model_output["reaction_projection"],
            model_output["sequence_projection"],
        )

        # Normalize
        z1 = nn.functional.normalize(z1, dim=1)
        z2 = nn.functional.normalize(z2, dim=1)

        # loss
        similarity_matrix_1 = (
            torch.einsum("ic, jc -> ij", z1, z1) / args.dclw_temperature
        )
        similarity_matrix_2 = (
            torch.einsum("ic, jc -> ij", z2, z2) / args.dclw_temperature
        )
        similarity_matrix_12 = (
            torch.einsum("ic, jc -> ij", z1, z2) / args.dclw_temperature
        )

        pos_terms = torch.diag(similarity_matrix_12)
        if args.use_von_mises_fisher:
            pos_loss = (-pos_terms * von_mises_fisher_w(pos_terms)).sum()
        else:
            pos_loss = -pos_terms.sum()

        neg_terms = [
            similarity_matrix_1.fill_diagonal_(0),
            similarity_matrix_2.fill_diagonal_(0),
            similarity_matrix_12.fill_diagonal_(0),
        ]
        neg_exp_terms = [torch.exp(n) for n in neg_terms]
        neg_loss = (
            torch.log(neg_exp_terms[0].sum(-1) + neg_exp_terms[2].sum(-1)).sum()
            + torch.log(neg_exp_terms[1].sum(-1) + neg_exp_terms[2].sum(-1)).sum()
        )

        loss = neg_loss + 2 * pos_loss

        l_dict["dclw_loss"] = args.dclw_loss_lambda * loss.detach()

        p_dict["reaction_hidden"] = model_output["reaction_hidden"].detach()
        p_dict["sequence_hidden"] = model_output["sequence_hidden"].detach()

        return args.dclw_loss_lambda * loss, l_dict, p_dict

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--dclw_loss_lambda",
            type=float,
            default=1.0,
            help="Lambda to weigh the cross-entropy loss.",
        )
        parser.add_argument(
            "--dclw_temperature",
            type=float,
            default=1.0,
            help="temperature for dcl-w loss.",
        )


@register_object("clip_loss", "loss")
class CLIPLoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        logging_dict, predictions = {}, {}
        substrate_features = model_output["substrate_hiddens"]
        protein_features = model_output["protein_hiddens"]

        # cosine similarity as logits
        logit_scale = model.model.logit_scale.exp()

        if (len(substrate_features.shape) == 2) and (len(protein_features.shape) == 2):
            if (args.clip_loss_use_gather) and (int(args.gpus) > 1):
                # gather
                substrate_features_all = concat_all_gather(substrate_features)
                protein_features_all = concat_all_gather(protein_features)

                # contrast against all molecules
                logits_per_protein = (
                    logit_scale * protein_features @ substrate_features_all.t()
                )
                # contrast against all proteins
                logits_per_substrate = (
                    logit_scale * substrate_features @ protein_features_all.t()
                )

                rank = dist.get_rank()
                batch_size = substrate_features.size(0)
                labels = torch.linspace(
                    rank * batch_size,
                    rank * batch_size + batch_size - 1,
                    batch_size,
                    dtype=int,
                ).to(substrate_features.device)

            else:
                logits_per_substrate = (
                    logit_scale * substrate_features @ protein_features.t()
                )
                logits_per_protein = logits_per_substrate.t()

                # labels
                labels = torch.arange(logits_per_substrate.shape[0]).to(
                    logits_per_substrate.device
                )

        elif len(protein_features.shape) == 3:
            if (args.clip_loss_use_gather) and (int(args.gpus) > 1):
                substrate_features_all = concat_all_gather(substrate_features)
                protein_features_all = concat_all_gather(protein_features)

                rank = dist.get_rank()
                batch_size = substrate_features.size(0)
                labels = torch.linspace(
                    rank * batch_size,
                    rank * batch_size + batch_size - 1,
                    batch_size,
                    dtype=int,
                ).to(substrate_features.device)
            else:
                substrate_features_all = substrate_features
                protein_features_all = protein_features
                labels = torch.arange(protein_features.shape[0]).to(
                    protein_features.device
                )

            logits_per_substrate = torch.matmul(
                substrate_features.unsqueeze(1).unsqueeze(1),
                protein_features_all.permute(0, 2, 1),
            ).squeeze()  # num graphs, num proteins, sequence length
            logits_per_substrate, _ = logits_per_substrate.max(-1)

            logits_per_protein = torch.matmul(
                protein_features.unsqueeze(1), substrate_features_all.unsqueeze(-1)
            ).squeeze()  # num proteins, num graphs, sequence length
            logits_per_protein, _ = logits_per_protein.max(-1)

            if len(logits_per_protein.shape) == 1:
                logits_per_protein = logits_per_protein.unsqueeze(0)
            if len(logits_per_substrate.shape) == 1:
                logits_per_substrate = logits_per_substrate.unsqueeze(0)

        else:
            raise NotImplementedError

        loss = (
            F.cross_entropy(
                logits_per_protein,
                labels,
                label_smoothing=args.clip_loss_label_smoothing,
            )
            + F.cross_entropy(
                logits_per_substrate,
                labels,
                label_smoothing=args.clip_loss_label_smoothing,
            )
        ) / 2

        logging_dict["clip_loss"] = loss.detach()

        probs = F.softmax(logits_per_substrate, dim=-1).detach()
        predictions["clip_probs"] = probs
        predictions["clip_preds"] = probs.argmax(axis=-1).reshape(-1)
        predictions["clip_golds"] = labels

        return loss, logging_dict, predictions

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--clip_loss_use_gather",
            action="store_true",
            default=False,
            help="whether to gather across gpus when computing loss.",
        )
        parser.add_argument(
            "--clip_loss_label_smoothing",
            type=float,
            default=0.0,
            help="label smoothing to use.",
        )


@register_object("supervised_clip_loss", "loss")
class SupervisedCLIPLoss(Nox):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self, model_output, batch, model, args):
        logging_dict, predictions = {}, {}
        substrate_features = model_output["substrate_hiddens"]
        protein_features = model_output["protein_hiddens"]

        mask = []
        for smiles in batch.smiles:
            mask.append([smiles in r for r in batch.all_reactants])

        mask_positives = torch.tensor(mask).float().to(substrate_features.device)

        # cosine similarity as logits
        logits = substrate_features @ protein_features.t() / args.supcon_temperature

        loss_prots = self.norm_and_compute_loss(logits, mask_positives, args)
        loss_substrates = self.norm_and_compute_loss(
            logits.t(), mask_positives.t(), args
        )

        loss = ((loss_prots + loss_substrates) / 2) * args.sup_clip_loss_lambda

        logging_dict["clip_loss"] = loss.detach()

        predictions["substrate_hiddens"] = substrate_features.detach()
        predictions["protein_hiddens"] = protein_features.detach()

        # logits substrates @ prots -> prots @ substrates
        logits = logits.t()
        probs = torch.sigmoid(logits.detach())
        if logits.shape[-1] != args.batch_size:
            logits = torch.cat(
                [
                    logits,
                    torch.zeros(
                        logits.shape[0],
                        args.batch_size - logits.shape[1],
                        device=logits.device,
                    ),
                ],
                dim=-1,
            )
            probs = torch.cat(
                [
                    probs,
                    torch.zeros(
                        probs.shape[0],
                        args.batch_size - probs.shape[1],
                        device=logits.device,
                    ),
                ],
                dim=-1,
            )
            mask_positives = torch.cat(
                [
                    mask_positives.t(),
                    torch.zeros(
                        mask_positives.t().shape[0],
                        args.batch_size - mask_positives.t().shape[1],
                        device=logits.device,
                    ),
                ],
                dim=-1,
            )
            predictions["golds"] = mask_positives
        else:
            predictions["golds"] = (
                mask_positives.t()
            )  # so that mask is prots @ substrates

        predictions["logits"] = logits
        predictions["probs"] = probs
        predictions["preds"] = (probs > 0.5).float()
        predictions["y"] = batch.y  # labels of pairs, for testing

        return loss, logging_dict, predictions

    def norm_and_compute_loss(self, logits, mask_positives, args):
        # for numerical stability
        logits_max, _ = torch.max(logits, dim=1, keepdim=True)
        eps = 1e-8  # to avoid log(0)
        logits = logits - logits_max.detach() + eps
        exp_logits = torch.exp(logits)

        positive_logits = mask_positives * logits  # mask out negatives

        # compute log_prob
        log_prob = (
            positive_logits + eps - torch.log(exp_logits.sum(1, keepdim=True) + eps)
        )

        # compute mean of log-likelihood over positives
        pos_samples = mask_positives.sum(dim=1)

        # loss_per_sample = -(log_prob * mask_combined).sum(dim=1) / pos_samples
        loss_per_sample = -(log_prob * mask_positives).sum(dim=1) / (pos_samples + eps)
        loss = loss_per_sample.mean()
        return loss

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--supcon_temperature",
            type=float,
            default=0.07,
            help="temperature for supervised contrastive loss.",
        )
        parser.add_argument(
            "--sup_clip_loss_lambda",
            type=float,
            default=1.0,
            help="",
        )



from typing import Optional, List, NamedTuple
from io import BytesIO
from zipfile import ZipFile
from urllib.request import urlopen
from pathlib import Path
import os
import torch
from rich import print
import pytorch_lightning as pl
import argparse
from clipzyme.models.protmol import EnzymeReactionCLIP

CHECKPOINT_URL = "https://github.com/pgmikhael/clipzyme/releases/download/v1.0.0/clipzyme_checkpoints.zip"  # TODO: Update this


def download_and_extract(remote_model_url: str, local_model_dir) -> List[str]:
    resp = urlopen(remote_model_url)
    os.makedirs(local_model_dir, exist_ok=True)
    with ZipFile(BytesIO(resp.read())) as zip_file:
        all_files_and_dirs = zip_file.namelist()
        zip_file.extractall(local_model_dir)
    assert len(all_files_and_dirs) == 1, "Expected only one file in the zip"
    return all_files_and_dirs[0]


def download_model(checkpoint_path) -> str:
    """
    Download trained clipzyme model.

    Parameters
    ----------
        checkpoint_path (str): path to where model should be located

    Returns
    -------
        str: local path to model
    """
    # Create cache folder if not exists
    cache = os.path.dirname(checkpoint_path)
    os.makedirs(cache, exist_ok=True)

    # Download model
    print(f"Downloading model to {cache}")
    filename = download_and_extract(CHECKPOINT_URL, cache)
    model_path = os.path.join(cache, filename)
    print(f"Model saved in {model_path}")
    return model_path


class CLIPZymeOutput(NamedTuple):
    scores: torch.Tensor
    protein_hiddens: torch.Tensor
    reaction_hiddens: torch.Tensor


class CLIPZyme(pl.LightningModule):
    def __init__(self, args: argparse.Namespace, device: Optional[str] = None) -> None:
        """
        Initialize a trained clipzyme model for inference.

        Parameters
        ----------
        checkpoint_path: str
            Path to a clipzyme checkpoint.
        device: str
            If provided, will run inference using this device.
            By default uses GPU, if available.
        """
        checkpoint_path = args.checkpoint_path
        # Check if path exists
        if (checkpoint_path is None) or (not os.path.exists(checkpoint_path)):
            try:
                # download model
                print(f"Model checkpoint not found at {checkpoint_path}")
                checkpoint_path = download_model(checkpoint_path)
            except Exception as e:
                raise FileNotFoundError(
                    f"Model checkpoint not found at {checkpoint_path} and failed to download model. {e}"
                )

        # Set device
        if device is not None:
            self.device = device
        elif self.device is not None:  # set in lightning module
            pass
        else:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"

        # Load model
        self.model = self.load_model(checkpoint_path)

        # prep directories
        self.prepare_directories(args)

    def prepare_directories(self, args: argparse.Namespace) -> None:
        """
        Prepare directories for saving hiddens and predictions.

        Parameters
        ----------
        args : argparse.Namespace
            Arguments from command line.
        """
        self.save_hiddens = args.save_hiddens
        self.save_predictions = args.save_predictions
        if args.save_hiddens or args.save_predictions:
            checkpoint_path = Path(args.checkpoint_path)
            dataset_path = Path(args.dataset_file_path)
            hiddens_dir = os.path.join(
                args.inference_dir, checkpoint_path.stem, dataset_path.stem
            )
            os.makedirs(hiddens_dir, exist_ok=True)
            self.hiddens_dir = Path(hiddens_dir)

        if args.save_predictions:
            self.predictions_dir = os.path.join(hiddens_dir, "predictions")
            os.makedirs(self.predictions_dir, exist_ok=True)
            self.predictions_dir = Path(self.predictions_dir)
            self.predictions_path = os.path.join(self.hiddens_dir, "predictions.csv")

    def load_model(self, path: str) -> EnzymeReactionCLIP:
        """
        Load model from path.

        Parameters
        ----------
        path : str
            Path to a clipzyme model checkpoint.

        Returns
        -------
        model
            Pretrained clipzyme model
        """
        # Load checkpoint
        checkpoint = torch.load(path, map_location="cpu")
        args = checkpoint["args"]
        model = EnzymeReactionCLIP(args)

        # Remove model from param names
        state_dict = {k[6:]: v for k, v in checkpoint["state_dict"].items()}
        model.load_state_dict(state_dict)  # type: ignore
        if self.device == "cuda":
            model.to("cuda")

        # Set eval
        model.eval()
        print(f"Loaded model from {path}")
        return model

    def forward(
        self,
        batch,
        batch_idx: int = 0,
        extract_protein_features: bool = False,
        extract_reaction_features: bool = False,
    ) -> CLIPZymeOutput:
        """
        Forward pass of the model.

        Parameters
        ----------
        batch : dict
            Batch of data.
        batch_idx : int, optional
            batch id, by default 0
        extract_protein_features : bool, optional
            whether to return protein features, by default False
        extract_reaction_features : bool, optional
            whether to return reaction features, by default False

        Returns
        -------
        CLIPZymeOutput
            Output of the model.
        """
        reaction_scores = None
        protein_hiddens = None
        reaction_hiddens = None

        model_output = self.model(batch)

        predictions_dict = {}
        predictions_dict = self.store_in_predictions(predictions_dict, batch)
        predictions_dict = self.store_in_predictions(predictions_dict, model_output)

        # want paired scores
        if not (self.use_as_protein_encoder or self.use_as_reaction_encoder):
            reaction_hiddens = model_output["substrate_hiddens"]
            protein_hiddens = model_output["protein_hiddens"]
            reaction_scores = torch.einsum(
                "bx,bx->b", reaction_hiddens, protein_hiddens
            )
        if extract_protein_features:
            protein_hiddens = (
                model_output["hidden"]
                if self.use_as_protein_encoder
                else model_output["protein_hiddens"]
            )
        if extract_reaction_features:
            reaction_hiddens = (
                model_output["hidden"]
                if self.use_as_protein_encoder
                else model_output["substrate_hiddens"]
            )

        output = CLIPZymeOutput(
            scores=reaction_scores,
            protein_hiddens=protein_hiddens,
            reaction_hiddens=reaction_hiddens,
            sample_ids=batch["sample_id"],
        )

        return output

    def test_step(self, batch, batch_idx) -> CLIPZymeOutput:
        """
        Test step for the model.

        Parameters
        ----------
        batch : dict
            Batch of data.
        batch_idx : int
            Batch index.

        Returns
        -------
        ClipZymeOutput
            Output of the model.
        """
        output = self.forward(batch, batch_idx)
        if self.save_hiddens or self.save_predictions:
            self.save_outputs(output)
        return output

    def on_test_epoch_end(self) -> None:
        self.clean_up()

    def save_outputs(self, outputs: CLIPZymeOutput) -> None:
        """
        Save outputs to disk.

        Parameters
        ----------
        outputs : ClipZymeOutput
            Output of the model.
        """
        if outputs.protein_hiddens is not None:
            protein_hiddens = outputs.protein_hiddens.cpu()
            for idx, protein_hidden in enumerate(protein_hiddens):
                predictions_filename = self.hiddens_dir.joinpath(
                    f"sample_{outputs.sample_ids[idx]}.protein.pt"
                )
                torch.save(protein_hidden, predictions_filename)

        if outputs.reaction_hidden is not None:
            reaction_hiddens = outputs.reaction_hidden.cpu()
            for idx, reaction_hidden in enumerate(reaction_hiddens):
                predictions_filename = self.hiddens_dir.joinpath(
                    f"sample_{outputs.sample_ids[idx]}.reaction.pt"
                )
                torch.save(reaction_hidden, predictions_filename)

        if outputs.scores is not None:
            scores = outputs.scores.cpu()
            for idx, score in enumerate(scores):
                predictions_filename = self.predictions_dir.joinpath(
                    f"sample_{outputs.sample_ids[idx]}.score.pt"
                )
                torch.save(score.item(), predictions_filename)

    def extract_protein_features(self, batch: dict) -> torch.Tensor:
        """
        Extract protein features from model.

        Parameters
        ----------
        batch : dict
            Batch of data.

        Returns
        -------
        torch.Tensor
            Protein features.
        """
        self.use_as_protein_encoder = True
        model_output = self.model(batch)
        return model_output["protein_hiddens"]

    def extract_reaction_features(self, batch: dict) -> torch.Tensor:
        """
        Extract reaction features from model.

        Parameters
        ----------
        batch : dict
            Batch of data.

        Returns
        -------
        torch.Tensor
            Reaction features.
        """
        self.use_as_reaction_encoder = True
        model_output = self.model(batch)
        return model_output["substrate_hiddens"]

    def clean_up(self):
        """
        Clean up directories after inference.
        """
        if self.save_predictions:
            print("Collecting predictions files")

            with open(self.predictions_path, "w") as f:
                f.write("sample_id,score\n")
                for score_file in self.predictions_dir.iterdir():
                    score = torch.load(self.predictions_dir.joinpath(score_file))
                    f.write(f"{score_file.stem},{score}\n")

            self.predictions_dir.rmdir()
        print("Done")
